#!/usr/bin/env python3
#
# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import distro
import sys

# access /usr/share/pyshared on Debian
# http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=715010

if distro.id() == 'debian':
    sys.path[0:0] = ['/usr/share/pyshared']

import grp
import pwd
import os
import socket

from dhcpy6d import UDPMulticastIPv6
from dhcpy6d.config import cfg

from dhcpy6d.globals import (config_answer_queue,
                             config_query_queue,
                             IF_NAME,
                             route_queue,
                             volatile_answer_queue,
                             volatile_query_queue)
from dhcpy6d.log import log
from dhcpy6d.handler import RequestHandler

from dhcpy6d.route import manage_prefixes_routes
from dhcpy6d.storage import (config_store,
                             QueryQueue,
                             volatile_store)
from dhcpy6d.threads import (DNSQueryThread,
                             RouteThread,
                             TidyUpThread,
                             TimerThread)


# main part, initializing all stuff
def run():
    log.info('Starting dhcpy6d daemon...')
    log.info(f'Server DUID: {cfg.SERVERDUID}')

    # configure SocketServer
    UDPMulticastIPv6.address_family = socket.AF_INET6
    udp_server = UDPMulticastIPv6(('', 547), RequestHandler)

    # start query queue watcher
    config_query_queue_watcher = QueryQueue(name='config_query_queue',
                                            store_type=config_store,
                                            query_queue=config_query_queue,
                                            answer_queue=config_answer_queue)
    config_query_queue_watcher.start()
    volatile_query_queue_watcher = QueryQueue(name='volatile_query_queue',
                                              store_type=volatile_store,
                                              query_queue=volatile_query_queue,
                                              answer_queue=volatile_answer_queue)
    volatile_query_queue_watcher.start()

    # check if database tables are up-to-date or exist and create them if not
    volatile_store.check_storage()

    # check if config database - if any - supports prefixes
    config_store.check_config_prefixes_support()

    # if global dynamic prefix was not given take it from database - only possible after database initialisation
    if cfg.PREFIX == '':
        cfg.PREFIX = volatile_store.get_dynamic_prefix()
    if cfg.PREFIX is None:
        cfg.PREFIX = ''

    # apply dynamic prefix to addresses and prefixes
    for a in cfg.ADDRESSES:
        cfg.ADDRESSES[a].inject_dynamic_prefix_into_prototype(cfg.PREFIX)
    for p in cfg.PREFIXES:
        cfg.PREFIXES[p].inject_dynamic_prefix_into_prototype(cfg.PREFIX)

    # collect all known MAC addresses from database
    if cfg.CACHE_MAC_LLIP:
        volatile_store.collect_macs_from_db()

    # start timer
    timer_thread = TimerThread()
    timer_thread.start()

    # start route queue to care for routes in background
    route_thread = RouteThread(route_queue)
    route_thread.start()

    # delete invalid and add valid routes - useful after reboot
    if cfg.MANAGE_ROUTES_AT_START:
        manage_prefixes_routes()

    # start TidyUp thread for cleaning in background
    tidyup_thread = TidyUpThread()
    tidyup_thread.start()

    # start DNS query queue to care for DNS in background
    dns_query_thread = DNSQueryThread()
    dns_query_thread.start()

    # set user and group
    log.info(f'Running as user {cfg.USER} (UID {pwd.getpwnam(cfg.USER).pw_uid}) and '
             f'group {cfg.GROUP} (GID {grp.getgrnam(cfg.GROUP).gr_gid})')
    # first set group because otherwise the freshly unprivileged user could not modify its groups itself
    os.setgid(grp.getgrnam(cfg.GROUP).gr_gid)
    os.setuid(pwd.getpwnam(cfg.USER).pw_uid)

    # log interfaces
    log.info(f'Listening on interfaces: {" ".join(IF_NAME)}')

    # serve forever
    try:
        udp_server.serve_forever()
    except KeyboardInterrupt:
        sys.exit(0)


if __name__ == '__main__':
    run()



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import grp
import os
import pwd
import sys
import traceback

from ..config import cfg
from .store import Store


class SQLite(Store):
    """
        file-based SQLite database, might be an option for single installations
    """
    QUERY_TABLES = "SELECT name FROM sqlite_master WHERE type='table'"

    def __init__(self, query_queue, answer_queue, storage_type='volatile'):

        Store.__init__(self, query_queue, answer_queue)
        self.connection = None

        try:
            self.db_connect(storage_type)
        except:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()

    def db_connect(self, storage_type='volatile'):
        """
        initialize DB connection
        """
        # only import if needed
        if 'sqlite3' not in sys.modules:
            import sqlite3
            self.db_module = sqlite3
        try:
            if storage_type == 'volatile':
                storage = cfg.STORE_SQLITE_VOLATILE
                # set ownership of storage file according to settings
                os.chown(cfg.STORE_SQLITE_VOLATILE, pwd.getpwnam(cfg.USER).pw_uid, grp.getgrnam(cfg.GROUP).gr_gid)
            if storage_type == 'config':
                storage = cfg.STORE_SQLITE_CONFIG
            self.connection = self.db_module.connect(storage, check_same_thread = False)
            self.cursor = self.connection.cursor()
            self.connected = True
        except:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()
            return None

    def db_query(self, query):
        """
        execute query on DB
        """
        try:
            self.cursor.execute(query)
            # commit only if explicitly wanted
            if query.startswith('INSERT'):
                self.connection.commit()
            elif query.startswith('UPDATE'):
                self.connection.commit()
            elif query.startswith('DELETE'):
                self.connection.commit()
            self.connected = True
        except self.db_module.IntegrityError:
            return 'INSERT_ERROR'
        except Exception as err:
            # try to reestablish database connection
            print(f'Error: {str(err.args[0])}')
            print(f'Query: {query}')
            if not self.db_connect():
                return None

        result = self.cursor.fetchall()
        return result



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re
import sys
import traceback

from ..config import cfg
from ..constants import CONST
from ..globals import (DUMMY_MAC,
                       EMPTY_OPTIONS,
                       IGNORED_LOG_OPTIONS)
from ..helpers import colonify_ip6
from ..log import log
from ..storage import config_store

from .default import default
from .from_config import from_config
from .parse_pattern import (parse_pattern_address,
                            parse_pattern_prefix)
from .reuse_lease import reuse_lease


class Client:
    """
        client object, generated from configuration database or on the fly
    """
    def __init__(self, transaction=None):
        # Addresses, depending on class or fixed addresses
        self.addresses = list()
        # Bootfiles, depending on class and architecture
        self.bootfiles = list()
        # Last chosen Bootfile
        self.chosen_boot_file = ''
        # Prefixes, depending on class or fixed prefixes
        self.prefixes = list()
        # DUID
        self.duid = ''
        # Hostname
        self.hostname = ''
        # Class/role of client - sadly "class" is a keyword and "class_" is more error prone
        self.client_class = ''
        # MAC
        self.mac = ''
        # timestamp of last update
        self.last_update = ''

        # when an transaction is given build the client
        if transaction is not None:
            self.build(transaction)

    def get_options_string(self):
        """
            all attributes in a string for logging
        """
        options_string = ''
        # put own attributes into a string
        options = sorted(self.__dict__.keys())
        # options.sort()
        for option in options:
            # ignore some attributes
            if option not in IGNORED_LOG_OPTIONS and self.__dict__[option] not in EMPTY_OPTIONS:
                if option == 'addresses':
                    if 'addresses' in cfg.CLASSES[self.client_class].ADVERTISE:
                        option_string = f'{option}:'
                        for address in self.__dict__[option]:
                            option_string += f' {colonify_ip6(address.ADDRESS)}'
                        options_string = f'{options_string} | {option_string}'
                elif option == 'bootfiles':
                    option_string = f'{option}:'
                    for bootfile in self.__dict__[option]:
                        option_string += f' {bootfile.BOOTFILE_URL}'
                    options_string = f'{options_string} | {option_string}'
                elif option == 'prefixes':
                    if 'prefixes' in cfg.CLASSES[self.client_class].ADVERTISE:
                        option_string = f'{option}:'
                        for p in self.__dict__[option]:
                            option_string += f' {colonify_ip6(p.PREFIX)}/{p.LENGTH}'
                        options_string = f'{options_string} | {option_string}'
                elif option == 'mac':
                    if self.__dict__[option] != DUMMY_MAC:
                        option_string = f'{option}: {self.__dict__[option]}'
                        options_string = f'{options_string} | {option_string}'
                else:
                    option_string = f'{option}: {self.__dict__[option]}'
                    options_string = f'{options_string} | {option_string}'
        return options_string

    def build(self, transaction):
        """
            builds client object of client config and transaction data
            checks if filters apply
            check if lease is still valid for RENEW and REBIND answers
            check if invalid addresses need to get deleted with lifetime 0
        """
        try:
            # configuration from client deriving from general config or filters - defaults to none
            client_config = None

            # list to collect filtered client information
            # if there are more than one entries that do not match the class is not uniquely identified
            filtered_class = {}

            # check if there are identification attributes of any class - classes are sorted by filter types
            for f in cfg.FILTERS:
                # look into all classes and their filters
                for c in cfg.FILTERS[f]:
                    # check further only if class applies to interface
                    if transaction.interface in c.INTERFACE:
                        # MACs
                        if c.FILTER_MAC != '':
                            pattern = re.compile(c.FILTER_MAC)
                            # if mac filter fits client mac address add client config
                            if len(pattern.findall(transaction.mac)) > 0:
                                client_config = config_store.get_client_config(hostname=transaction.hostname,
                                                                               mac=[transaction.mac],
                                                                               duid=transaction.duid,
                                                                               client_class=c.NAME)
                                # add classname to dictionary - if there are more than one entry classes do not match
                                # and thus are invalid
                                filtered_class[c.NAME] = c
                        # DUIDs
                        if c.FILTER_DUID != '':
                            pattern = re.compile(c.FILTER_DUID)
                            # if duid filter fits client duid address add client config
                            if len(pattern.findall(transaction.duid)) > 0:
                                client_config = config_store.get_client_config(hostname=transaction.hostname,
                                                                               mac=[transaction.mac],
                                                                               duid=transaction.duid,
                                                                               client_class=c.NAME)
                                # see above
                                filtered_class[c.NAME] = c
                        # HOSTNAMEs
                        if c.FILTER_HOSTNAME != '':
                            pattern = re.compile(c.FILTER_HOSTNAME)
                            # if hostname filter fits client hostname address add client config
                            if len(pattern.findall(transaction.hostname)) > 0:
                                client_config = config_store.get_client_config(hostname=transaction.hostname,
                                                                               mac=[transaction.mac],
                                                                               duid=transaction.duid,
                                                                               client_class=c.NAME)
                                # see above
                                filtered_class[c.NAME] = c

            # if there are more than 1 different classes matching for the client they are not valid
            if len(filtered_class) != 1:
                client_config = None

            # if filters did not get a result try it the hard way
            if client_config is None:
                # check all given identification criteria - if they all match each other the client is identified
                id_attributes = []

                # get client config that most probably seems to fit
                # only ask DB if store is a DB
                if cfg.STORE_CONFIG != 'file':
                    config_store.build_config_from_db(transaction)

                # check every attribute which is required
                # depending on identificaton mode empty results are ignored or considered
                # finally all attributes are grouped in sets and for a correctly identified host
                # only one entry should appear at the end
                for identification in cfg.IDENTIFICATION:
                    if identification == 'mac':
                        # get all MACs for client from config
                        macs = config_store.get_client_config_by_mac(transaction)
                        if macs:
                            macs = set(macs)
                            id_attributes.append(macs)
                        elif cfg.IDENTIFICATION_MODE == 'match_all':
                            macs = set()
                            id_attributes.append(macs)

                    if identification == 'duid':
                        duids = config_store.get_client_config_by_duid(transaction)
                        if duids:
                            duids = set(duids)
                            id_attributes.append(duids)
                        elif cfg.IDENTIFICATION_MODE == 'match_all':
                            duids = set()
                            id_attributes.append(duids)

                    if identification == 'hostname':
                        hostnames = config_store.get_client_config_by_hostname(transaction)
                        if hostnames:
                            hostnames = set(hostnames)
                            id_attributes.append(hostnames)
                        elif cfg.IDENTIFICATION_MODE == 'match_all':
                            hostnames = set()
                            id_attributes.append(hostnames)

                # get intersection of all sets of identifying attributes - even the empty ones
                if len(id_attributes) > 0:
                    client_config = set.intersection(*id_attributes)

                    # if exactly one client has been identified use that config
                    if len(client_config) == 1:
                        # reuse client_config, grab it out of the set
                        client_config = client_config.pop()
                    else:
                        # in case there is no client config we should maybe log this?
                        client_config = None
                else:
                    client_config = None

            # If client gave some addresses previously for RENEW or REBIND consider them
            if (transaction.last_message_received_type is CONST.MESSAGE.RENEW or
                transaction.last_message_received_type is CONST.MESSAGE.REBIND) and \
               not (len(transaction.addresses) == 0 and
                    len(transaction.prefixes) == 0):
                # use already existing lease
                reuse_lease(client=self, client_config=client_config, transaction=transaction)
            # build IA addresses from config - fixed ones and dynamic
            elif client_config is not None:
                # build client from config
                from_config(client=self, client_config=client_config, transaction=transaction)
            else:
                # use default class if host is unknown
                default(client=self, client_config=client_config, transaction=transaction)

        except Exception as err:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()
            log.error('build(): ' + str(err))
            return None



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re

from dhcpy6d.config import cfg
from dhcpy6d.helpers import (convert_binary_to_dns,
                             convert_dns_to_binary)
from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    Option 39 FQDN

    http://tools.ietf.org/html/rfc4704#page-5
    regarding RFC 4704 5. there are 3 kinds of client behaviour for N O S:
      - client wants to update DNS itself -> sends 0 0 0
      - client wants server to update DNS -> sends 0 0 1
       - client wants no server DNS update -> sends 1 0 0
    """

    def build(self, transaction=None, **kwargs):
        # dummy empty return value
        response_string_part = ''
        options_answer_part = None

        # http://tools.ietf.org/html/rfc4704#page-5
        # regarding RFC 4704 5. there are 3 kinds of client behaviour for N O S:
        # - client wants to update DNS itself -> sends 0 0 0
        # - client wants server to update DNS -> sends 0 0 1
        # - client wants no server DNS update -> sends 1 0 0
        if transaction.client:
            # flags for answer
            n, o, s = 0, 0, 0
            # use hostname supplied by client
            if cfg.DNS_USE_CLIENT_HOSTNAME:
                hostname = transaction.hostname
            # use hostname from config
            else:
                hostname = transaction.client.hostname
            if not hostname == '':
                if cfg.DNS_UPDATE:
                    # DNS update done by server - don't care what client wants
                    if cfg.DNS_IGNORE_CLIENT:
                        s = 1
                        o = 1
                    else:
                        # honor the client's request for the server to initiate DNS updates
                        if transaction.dns_s == 1:
                            s = 1
                        # honor the client's request for no server-initiated DNS update
                        elif transaction.dns_n == 1:
                            n = 1
                else:
                    # no DNS update at all, not for server and not for client
                    if transaction.dns_n == 1 or \
                            transaction.dns_s == 1:
                        o = 1
                # sum of flags
                nos_flags = n * 4 + o * 2 + s * 1
                fqdn_binary = convert_dns_to_binary(f'{hostname}.{cfg.DOMAIN}')
                response_string_part = self.convert_to_string(self.number, f'{nos_flags:02x}{fqdn_binary}')
            else:
                # if no hostname given put something in and force client override
                fqdn_binary = convert_dns_to_binary(f'invalid-hostname.{cfg.DOMAIN}')
                response_string_part = self.convert_to_string(self.number, f'{3:02x}{fqdn_binary}')
            # options in answer to be logged
            options_answer_part = self.number

        return response_string_part, options_answer_part

    def initialize(self, transaction=None, option=None, **kwargs):
        bits = f'{int(option[1:2]):04d}'
        transaction.dns_n = int(bits[1])
        transaction.dns_o = int(bits[2])
        transaction.dns_s = int(bits[3])
        # only hostname needed
        transaction.fqdn = convert_binary_to_dns(option[2:])
        transaction.hostname = transaction.fqdn.split('.')[0].lower()
        # test if hostname is valid
        hostname_pattern = re.compile('^([a-z0-9-_]+)*$')
        if hostname_pattern.match(transaction.hostname) is None:
            transaction.hostname = ''
        del hostname_pattern



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import configparser
import copy
import getopt
import grp
import os
import os.path
import platform
import pwd
import re
import shlex
import stat
import sys
import time
import uuid

from .helpers import (decompress_ip6,
                      error_exit,
                      get_interfaces,
                      listify_option,
                      LOCALHOST_INTERFACES,
                      send_control_message)

# needed for boolean options
BOOLPOOL = {'0': False, '1': True, 'no': False, 'yes': True, 'false': False, 'true': True, False: False, True: True,
            'on': True, 'off': False}

# whitespace for options with more than one value
WHITESPACE = ' ,'

# empty default prefix - if needed given by command line argument
PREFIX = ''

# default usage text - to be extended
USAGE = '''
dhcpy6d - DHCPv6 server

Usage: dhcpy6d --config <file> [--user <user>] [--group <group>] [--duid <duid>] [--prefix <prefix>] [--really-do-it <yes>|<no>]

       dhcpy6d --message '<message>'

       dhcpy6d --generate-duid

See manpage dhcpy6d(8) for details.
'''


class Config:
    """
      general settings
    """

    def __init__(self):
        """
            define defaults
        """
        # access dynamic PREFIX
        global PREFIX
        self.PREFIX = PREFIX

        # default settings
        # Server cfg.INTERFACE + addresses
        self.INTERFACE = ''
        self.ADDRESS = '2001:db8::1'
        # issue #50 proposes an except-interface option like in dnsmasq - interfaces not to listen at
        self.EXCLUDE_INTERFACE = ''
        # effective user and group - will have to be set mainly by distribution package
        self.USER = 'root'
        self.GROUP = 'root'
        # lets make the water turn black... or build a shiny server DUID
        # in case someone will ever debug something here: Wireshark shows
        # year 2042 even if it is 2012 - time itself is OK
        self.SERVERDUID = generate_duid()
        self.NAMESERVER = ''

        # domain for FQDN hostnames
        self.DOMAIN = 'domain'
        # domain search list for option 24, according to RFC 3646
        # defaults to DOMAIN
        self.DOMAIN_SEARCH_LIST = ''

        # IA_NA Options
        # Default preferred lifetime for addresses
        self.PREFERRED_LIFETIME = '5400'
        # Default valid lifetime for addresses
        self.VALID_LIFETIME = '7200'
        # T1 RENEW
        self.T1 = '2700'
        # T2 REBIND
        self.T2 = '4050'

        # Server Preference
        self.SERVER_PREFERENCE = '255'

        # SNTP SERVERS Option 31
        self.SNTP_SERVERS = ''

        # NTP server Option 56
        self.NTP_SERVER = ''
        # Auxiliary options, derived from self.NTP_SERVER
        self.NTP_SERVER_DICT = {'SRV': [], 'MC': [], 'FQDN': []}

        # INFORMATION REFRESH TIME option 32 for option 11 (INFORMATION REQUEST)
        # see RFC http://tools.ietf.org/html/rfc4242
        self.INFORMATION_REFRESH_TIME = '6000'

        # config type
        # one of file, mysql, sqlite or none
        self.STORE_CONFIG = 'none'
        # one of mysql, postgresql or sqlite
        self.STORE_VOLATILE = 'sqlite'

        # file for client information
        self.STORE_FILE_CONFIG = '/etc/dhcpy6d-clients.conf'

        # DB data
        self.STORE_DB_HOST = 'localhost'
        self.STORE_DB_DB = 'dhcpy6d'
        self.STORE_DB_USER = 'user'
        self.STORE_DB_PASSWORD = 'password'

        self.STORE_SQLITE_CONFIG = 'config.sqlite'
        self.STORE_SQLITE_VOLATILE = '/var/lib/dhcpy6d/volatile.sqlite'

        # whether MAC-LLIP pairs should be stored forever or retrieved freshly if needed
        self.CACHE_MAC_LLIP = 'False'

        # DNS Update settings
        self.DNS_UPDATE = 'False'
        self.DNS_UPDATE_NAMESERVER = '::1'
        self.DNS_TTL = 86400
        self.DNS_USE_RNDC = 'True'
        self.DNS_RNDC_KEY = 'rndc-key'
        self.DNS_RNDC_SECRET = '0000000000000000000000000000000000000000000000000000000000000'
        # DNS RFC 4704 client DNS wishes
        # use client supplied hostname
        self.DNS_USE_CLIENT_HOSTNAME = 'False'
        # ignore client ideas about DNS (if at all, what name to use, self-updating...)
        self.DNS_IGNORE_CLIENT = 'True'

        # Log ot not
        self.LOG = 'False'
        # Log level
        self.LOG_LEVEL = 'INFO'
        # Log on console
        self.LOG_CONSOLE = 'False'
        # Logfile
        self.LOG_FILE = ''
        # Log to syslog
        self.LOG_SYSLOG = 'False'
        # Syslog facility
        self.LOG_SYSLOG_FACILITY = 'daemon'
        # Local syslog socket or server:port
        if platform.system() in ['Linux', 'OpenBSD']:
            self.LOG_SYSLOG_DESTINATION = '/dev/log'
        else:
            self.LOG_SYSLOG_DESTINATION = '/var/run/log'

        # log newly found MAC addresses - if CACHE_MAC_LLIP is false this might be way too much
        self.LOG_MAC_LLIP = 'False'

        # some 128 bits
        self.AUTHENTICATION_INFORMATION = '00000000000000000000000000000000'

        # for debugging - if False nothing is done
        self.REALLY_DO_IT = 'False'

        # interval for TidyUp thread - time to sleep in TidyUpThread
        self.CLEANING_INTERVAL = 10

        # address, bootfile and class schemes
        self.ADDRESSES = {}
        self.BOOTFILES = {}
        self.CLASSES = {}
        self.PREFIXES = {}

        # how to identify clients
        self.IDENTIFICATION = 'mac'
        self.IDENTIFICATION_MODE = 'match_all'

        # allow one to ignore IAIDs which play no big role at all for server
        self.IGNORE_IAID = 'False'

        # ignore clients which do no appear in the neighbor cache table
        self.IGNORE_UNKNOWN_CLIENTS = 'True'

        # ignore MAC addresses as identifier - useful for neighbor-cache-less interfaces like ppp0
        self.IGNORE_MAC = 'False'

        # ignore interface to be able to listen on dynamically created interfaces like ppp
        self.IGNORE_INTERFACE = 'False'

        # allow setting request rate limits to put clients onto blacklist
        self.REQUEST_LIMIT = 'no'
        self.REQUEST_LIMIT_TIME = '60'
        self.REQUEST_LIMIT_COUNT = '20'
        self.REQUEST_LIMIT_RELEASE_TIME = '7200'
        self.REQUEST_LIMIT_IDENTIFICATION = 'llip'

        # restore still valid routes at startup and remove inactive ones
        self.MANAGE_ROUTES_AT_START = 'False'

        # regexp filters for hostnames etc.
        self.FILTERS = {'mac': [], 'duid': [], 'hostname': []}

        # define a fallback default class and address scheme
        self.ADDRESSES['default'] = Address(ia_type='na',
                                            category='mac',
                                            pattern='fdef::$mac$',
                                            aclass='default',
                                            atype='default',
                                            prototype='fdef0000000000000000xxxxxxxxxxxx')

        # define dummy address scheme for fixed addresses
        # pattern and prototype are not really needed as this
        # addresses are fixed
        self.ADDRESSES['fixed'] = Address(ia_type='na',
                                          category='fixed',
                                          pattern='fdef::1',
                                          aclass='default',
                                          atype='fixed',
                                          prototype='fdef0000000000000000000000000000')

        self.PREFIXES['default'] = Prefix(pattern='fdef:$range$::',
                                          prange='1000-1fff',
                                          category='range')

        self.CLASSES['default'] = Class()
        self.CLASSES['default'].ADDRESSES.append('default')
        self.CLASSES['default'].PREFIXES.append('default')

        # config file from command line
        # default config file and cli values
        configfile = self.cli_options = self.cli_user = self.cli_group = self.cli_duid = self.cli_really_do_it = None
        # get multiple options
        try:
            self.cli_options, cli_remains = getopt.gnu_getopt(sys.argv[1:],
                                                              'c:u:g:d:r:p:m:G',
                                                              ['config=',
                                                               'user=',
                                                               'group=',
                                                               'duid=',
                                                               'really-do-it=',
                                                               'prefix=',
                                                               'message=',
                                                               'generate-duid'])
            for opt, arg in self.cli_options:
                if opt in ('-c', '--config'):
                    configfile = arg
                if opt in ('-g', '--group'):
                    self.cli_group = arg
                if opt in ('-u', '--user'):
                    self.cli_user = arg
                if opt in ('-d', '--duid'):
                    self.cli_duid = arg
                if opt in ('-r', '--really-do-it'):
                    self.cli_really_do_it = arg
                if opt in ('-p', '--prefix'):
                    PREFIX = arg
                    self.PREFIX = PREFIX
                if opt in ('-m', '--message'):
                    send_control_message(arg)
                    sys.exit(0)
                if opt in ('-G', '--generate-duid'):
                    print(generate_duid())
                    sys.exit(0)

        except getopt.GetoptError as err:
            print(err)
            print(USAGE)
            sys.exit(0)

        if configfile is None:
            error_exit('No config file given - please use --config <config.file>')

        if os.path.exists(configfile):
            if not (os.path.isfile(configfile) or
                    os.path.islink(configfile)):
                error_exit(f"Configuration file '{configfile}' is no file or link.")
        else:
            error_exit(f"Configuration file '{configfile}' does not exist.")

        # read config at once
        self.read_config(configfile)

    def read_config(self, configfile):
        """
            read configuration from file, should work with included files too - at least this is the plan
        """

        # instantiate Configparser
        config = configparser.ConfigParser()
        config.read(configfile)

        # error message prefix for check config some lines later
        msg_prefix = 'General configuration:'

        # whyever sections classes get overwritten sometimes and so some configs had been missing
        # so create classes and addresses here
        for section in config.sections():
            # global PXE boot url schemes
            if section.startswith('bootfile_'):
                self.BOOTFILES[section.split('bootfile_')[1]] = BootFile(name=section.split('bootfile_')[1].strip())
            if section.startswith('class_'):
                self.CLASSES[section.split('class_')[1]] = Class(name=section.split('class_')[1].strip())
            if section.startswith('address_'):
                self.ADDRESSES[section.split('address_')[1].strip()] = Address()
            if section.startswith('prefix_'):
                self.PREFIXES[section.split('prefix_')[1].strip()] = Prefix()

        for section in config.sections():
            # go through all items
            for item in config.items(section):
                if section.upper() == 'DHCPY6D':
                    # check for legacy settings - STORE_MYSQL_* will be replaced by STORE_DB_* since 0.4.2
                    # see https://github.com/HenriWahl/dhcpy6d/issues/3
                    if item[0].upper() in ('STORE_MYSQL_HOST',
                                           'STORE_MYSQL_DB',
                                           'STORE_MYSQL_USER',
                                           'STORE_MYSQL_PASSWORD'):
                        sys.stderr.write(f"\nWARNING: Keyword '{item[0]}' in section '[{section}]' "
                                         f"is deprecated and should be replaced "
                                         f"by '{item[0].lower().replace('mysql', 'db')}'.\n\n")
                        object.__setattr__(self, item[0].upper().replace('MYSQL', 'DB'), str(item[1]).strip())
                    # keyword is deprecated
                    elif item[0].upper() in ('MCAST',
                                             'PORT'):
                        sys.stderr.write(f"\nWARNING: Keyword '{item[0]}' in section '[{section}]' "
                                         f"is deprecated and can safely be removed.\n")
                    # check if keyword is known - if not, exit
                    elif not self.__dict__:
                        error_exit(f"Keyword '{item[0]}' in section '[{section}]' "
                                   f"of configuration file '{configfile}' is unknown.")
                    # ConfigParser seems to be not case sensitive so settings get normalized
                    else:
                        object.__setattr__(self, item[0].upper(), str(item[1]).strip())

                        # treat interface exceptions here already to be able to use non-excluded interfaces in other sections
                        # to be able to set used interfaces the except_interfaces has to be evaluated here
                        if item[0].upper() in ['INTERFACE', 'EXCLUDE_INTERFACE']:
                            # get interfaces as list
                            self.INTERFACE = listify_option(self.INTERFACE)
                            self.EXCLUDE_INTERFACE = listify_option(self.EXCLUDE_INTERFACE)
                            if self.EXCLUDE_INTERFACE:
                                # just make sure there is either interface or except_interface set
                                if self.INTERFACE:
                                    error_exit(f"{msg_prefix} Interface and excluded interface are mutually exclusive.")
                                else:
                                    # check if interface is not excluded and not localhost interface
                                    for interface in self.EXCLUDE_INTERFACE:
                                        if not interface in get_interfaces():
                                            error_exit(f"{msg_prefix} Excluded interface '{interface}' is unknown.")

                                # reset interface to be filled with non-excluded interfaces
                                self.INTERFACE = []
                                # set non-excluded interfaces
                                for interface in get_interfaces():
                                    if not interface in self.EXCLUDE_INTERFACE and \
                                            not interface in LOCALHOST_INTERFACES:
                                        self.INTERFACE.append(interface)
                else:
                    # global PXE boot url schemes
                    if section.lower().startswith('bootfile_'):
                        if not item[0].upper() in self.BOOTFILES[section.lower().split('bootfile_')[1]].__dict__:
                            error_exit(f"Keyword '{item[0]}' in section '[{section}]' "
                                       f"of configuration file '{configfile}' is unknown.")
                        self.BOOTFILES[section.lower().split('bootfile_')[1]].__setattr__(item[0].upper(),
                                                                                          str(item[1]).strip())
                    # global address schemes
                    if section.lower().startswith('address_'):
                        # check if keyword is known - if not, exit
                        if item[0].upper() == 'PREFIX_LENGTH':
                            # Show a warning because there are no prefix lengths in DHCPv6
                            sys.stderr.write(f"\nWARNING: Keyword '{item[0]}' in section '{section}' is deprecated "
                                             "and should be removed.\n\n")
                        else:
                            if not item[0].upper() in self.ADDRESSES[section.lower().split('address_')[1]].__dict__:
                                error_exit(f"Keyword '{item[0]}' in section '[{section}]' "
                                           f"of configuration file '{configfile}' is unknown.")
                        self.ADDRESSES[section.lower().split('address_')[1]].__setattr__(item[0].upper(),
                                                                                         str(item[1]).strip())

                    # global prefix schemes
                    if section.lower().startswith('prefix_'):
                        if not item[0].upper() in self.PREFIXES[section.lower().split('prefix_')[1]].__dict__:
                            error_exit(f"Keyword '{item[0]}' in section '[{section}]' "
                                       f"of configuration file '{configfile}' is unknown.")
                        self.PREFIXES[section.lower().split('prefix_')[1]].__setattr__(item[0].upper(),
                                                                                       str(item[1]).strip())

                    # global classes with their addresses
                    elif section.lower().startswith('class_'):
                        # check if keyword is known - if not, exit
                        if not item[0].upper() in self.CLASSES[section.lower().split('class_')[1]].__dict__:
                            error_exit(f"Keyword '{item[0]}' in section '[{section}]' "
                                       f"of configuration file '{configfile}' is unknown.")
                        if item[0].upper() == 'ADDRESSES':
                            # strip whitespace and separators of addresses
                            lex = shlex.shlex(item[1])
                            lex.whitespace = WHITESPACE
                            lex.wordchars += ':.'
                            for address in lex:
                                if len(address) > 0:
                                    self.CLASSES[section.lower().split('class_')[1]].ADDRESSES.append(address)
                        elif item[0].upper() == 'BOOTFILES':
                            # strip whitespace and separators of bootfiles
                            lex = shlex.shlex(item[1])
                            lex.whitespace = WHITESPACE
                            lex.wordchars += ':.'
                            for bootfile in lex:
                                if len(bootfile) > 0:
                                    self.CLASSES[section.lower().split('class_')[1]].BOOTFILES.append(bootfile)
                        elif item[0].upper() == 'PREFIXES':
                            # strip whitespace and separators of prefixes
                            lex = shlex.shlex(item[1])
                            lex.whitespace = WHITESPACE
                            lex.wordchars += ':.'
                            for prefix in lex:
                                if len(prefix) > 0:
                                    self.CLASSES[section.lower().split('class_')[1]].PREFIXES.append(prefix)
                        elif item[0].upper() == 'ADVERTISE':
                            # strip whitespace and separators of advertised IAs
                            lex = shlex.shlex(item[1])
                            lex.whitespace = WHITESPACE
                            lex.wordchars += ':.'
                            self.CLASSES[section.lower().split('class_')[1]].ADVERTISE[:] = []
                            for advertise in lex:
                                if len(advertise) > 0:
                                    self.CLASSES[section.lower().split('class_')[1]].ADVERTISE.append(advertise)
                        elif item[0].upper() == 'INTERFACE':
                            # strip whitespace and separators of interfaces
                            lex = shlex.shlex(item[1])
                            lex.whitespace = WHITESPACE
                            lex.wordchars += ':.'
                            for interface in lex:
                                if interface not in self.INTERFACE:
                                    error_exit(f"Interface '{interface}' used in section '[{section}]' "
                                               f"of configuration file '{configfile}' is not "
                                               "defined in general settings.")
                                self.CLASSES[section.lower().split('class_')[1]].INTERFACE.append(interface)
                        else:
                            self.CLASSES[section.lower().split('class_')[1]].__setattr__(item[0].upper(),
                                                                                         str(item[1]).strip())

        # The next paragraphs contain finetuning
        self.IDENTIFICATION = listify_option(self.IDENTIFICATION)

        # at least something interfacy should be configured
        if not self.INTERFACE and \
                not self.EXCLUDE_INTERFACE:
            error_exit(f"{msg_prefix} Neither interface or excluded interface is defined.")

        # create default classes for each interface - if not defined
        # derive from default 'default' class
        for interface in self.INTERFACE:
            if not 'default_' + interface in self.CLASSES:
                self.CLASSES['default_' + interface] = copy.copy(self.CLASSES['default'])
                self.CLASSES['default_' + interface].NAME = 'default_' + interface
                self.CLASSES['default_' + interface].INTERFACE = [interface]

        # lower storage
        self.STORE_CONFIG = self.STORE_CONFIG.lower()
        self.STORE_VOLATILE = self.STORE_VOLATILE.lower()

        # boolize none-config-store
        if self.STORE_CONFIG.lower() == 'none':
            self.STORE_CONFIG = False

        # if no domain search list has been given use DOMAIN
        if len(self.DOMAIN_SEARCH_LIST) == 0:
            self.DOMAIN_SEARCH_LIST = self.DOMAIN

        # domain search list has to be a list
        self.DOMAIN_SEARCH_LIST = listify_option(self.DOMAIN_SEARCH_LIST)

        # get nameservers as list
        if len(self.NAMESERVER) > 0:
            self.NAMESERVER = listify_option(self.NAMESERVER)

        # option 31 quite probably is obsolete but might still be used, so just take its values from newer option 56
        # client dhcpcd for example uses this option when asking for NTP server
        if len(self.SNTP_SERVERS) > 0:
            self.SNTP_SERVERS = listify_option(self.SNTP_SERVERS)
        elif len(self.NTP_SERVER) > 0:
            self.SNTP_SERVERS = listify_option(self.NTP_SERVER)

        # get NTP servers as list
        if len(self.NTP_SERVER) > 0:
            self.NTP_SERVER = listify_option(self.NTP_SERVER)

        # convert to boolean values
        for option in ['DNS_IGNORE_CLIENT',
                       'DNS_USE_CLIENT_HOSTNAME',
                       'DNS_USE_RNDC',
                       'DNS_UPDATE',
                       'REALLY_DO_IT',
                       'LOG',
                       'LOG_CONSOLE',
                       'LOG_SYSLOG',
                       'CACHE_MAC_LLIP',
                       'LOG_MAC_LLIP',
                       'IGNORE_IAID',
                       'IGNORE_UNKNOWN_CLIENTS',
                       'IGNORE_MAC',
                       'IGNORE_INTERFACE',
                       'REQUEST_LIMIT',
                       'MANAGE_ROUTES_AT_START']:
            try:
                self.__dict__[option] = BOOLPOOL[self.__dict__[option].lower()]
            except:
                error_exit(f"Option '{option.lower()}' only allows boolean values like 'yes' and 'no'.")

        # upperize for syslog
        self.LOG_SYSLOG_FACILITY = self.LOG_SYSLOG_FACILITY.upper()
        self.LOG_LEVEL = self.LOG_LEVEL.upper()

        # index of classes which add some identification rules etc.
        for c in list(self.CLASSES.values()):
            if c.FILTER_MAC != '':
                self.FILTERS['mac'].append(c)
            if c.FILTER_DUID != '':
                self.FILTERS['duid'].append(c)
            if c.FILTER_HOSTNAME != '':
                self.FILTERS['hostname'].append(c)
            if c.NAMESERVER != '':
                c.NAMESERVER = listify_option(c.NAMESERVER)
            if c.NTP_SERVER != '':
                c.NTP_SERVER = listify_option(c.NTP_SERVER)
            if len(c.INTERFACE) == 0:
                # use general setting if none specified
                c.INTERFACE = self.INTERFACE
            # use default T1 and T2 if not defined
            if c.T1 == 0:
                c.T1 = self.T1
            if c.T2 == 0:
                c.T2 = self.T2
            # check advertised IA types - if empty default to ['addresses']
            if len(c.ADVERTISE) == 0:
                c.ADVERTISE = ['addresses', 'prefixes']

        # set type properties for addresses
        for a in self.ADDRESSES:
            # name for address, important for leases db
            self.ADDRESSES[a].TYPE = a
            if self.ADDRESSES[a].VALID_LIFETIME == 0:
                self.ADDRESSES[a].VALID_LIFETIME = self.VALID_LIFETIME
            if self.ADDRESSES[a].PREFERRED_LIFETIME == 0:
                self.ADDRESSES[a].PREFERRED_LIFETIME = self.PREFERRED_LIFETIME
            # normalize ranges
            self.ADDRESSES[a].RANGE = self.ADDRESSES[a].RANGE.lower()
            # convert boolean string to boolean value
            self.ADDRESSES[a].DNS_UPDATE = BOOLPOOL[self.ADDRESSES[a].DNS_UPDATE]
            if self.ADDRESSES[a].DNS_ZONE == '':
                self.ADDRESSES[a].DNS_ZONE = self.DOMAIN
            if self.ADDRESSES[a].DNS_TTL == '0':
                self.ADDRESSES[a].DNS_TTL = self.DNS_TTL
            # add prototype for later fast validity comparison of rebinding leases
            # also use as proof of validity of address patterns
            self.ADDRESSES[a].build_prototype()

        # set type properties for prefixes
        for p in self.PREFIXES:
            # name for address, important for leases db
            self.PREFIXES[p].TYPE = p
            if self.PREFIXES[p].VALID_LIFETIME == 0:
                self.PREFIXES[p].VALID_LIFETIME = self.VALID_LIFETIME
            if self.PREFIXES[p].PREFERRED_LIFETIME == 0:
                self.PREFIXES[p].PREFERRED_LIFETIME = self.PREFERRED_LIFETIME
            # normalize ranges
            self.PREFIXES[p].RANGE = self.PREFIXES[p].RANGE.lower()
            # route via Link Local Address
            self.PREFIXES[p].ROUTE_LINK_LOCAL = BOOLPOOL[self.PREFIXES[p].ROUTE_LINK_LOCAL]
            # add prototype for later fast validity comparison of rebinding leases
            # also use as proof of validity of address patterns
            self.PREFIXES[p].build_prototype()

        # check if some options are set by cli options
        if self.cli_user is not None:
            self.USER = self.cli_user
        if not self.cli_group is None:
            self.GROUP = self.cli_group
        if not self.cli_duid is None:
            self.SERVERDUID = self.cli_duid
        if not self.cli_really_do_it is None:
            self.REALLY_DO_IT = BOOLPOOL[self.cli_really_do_it.lower()]

        # check user and group
        try:
            pwd.getpwnam(self.USER)
        except:
            error_exit(f"{msg_prefix} User '{self.USER}' does not exist")
        try:
            grp.getgrnam(self.GROUP)
        except:
            error_exit(f"{msg_prefix} Group '{self.GROUP}' does not exist")

        # check interface
        if not self.IGNORE_INTERFACE:
            for i in self.INTERFACE:
                # also accept Linux VLAN and other definitions but interface must exist
                if not i in get_interfaces() or not re.match('^[a-z0-9_:.%-]*$', i, re.IGNORECASE):
                    error_exit(f"{msg_prefix} Interface '{i}' is unknown.")

        # check server's address
        try:
            decompress_ip6(self.ADDRESS)
        except Exception as err:
            error_exit(f"{msg_prefix} Server address '{err}' is invalid.")

        # check server duid
        if not self.SERVERDUID.isalnum():
            error_exit(f"{msg_prefix} Server DUID '{self.SERVERDUID}' must be alphanumeric.")

        # check nameserver to be given to client
        for nameserver in self.NAMESERVER:
            try:
                decompress_ip6(nameserver)
            except Exception as err:
                error_exit(f"{msg_prefix} Name server address '{err}' is invalid.")

        # split NTP server types into possible 3 (address, multicast, FQDN)
        # more details about this madness are available at https://tools.ietf.org/html/rfc5908
        for ntp_server in self.NTP_SERVER:
            try:
                decompress_ip6(ntp_server)
                # if decompressing worked it must be an address
                if ntp_server.lower().startswith('ff'):
                    self.NTP_SERVER_DICT['MC'].append(ntp_server.lower())
                else:
                    self.NTP_SERVER_DICT['SRV'].append(ntp_server.lower())
            except Exception as err:
                if re.match('^[a-z0-9.-]*$', ntp_server, re.IGNORECASE):
                    self.NTP_SERVER_DICT['FQDN'].append(ntp_server.lower())
                else:
                    error_exit(f"{msg_prefix} NTP server address '{ntp_server}' is invalid.")

        # partly check of domain name validity
        if not re.match('^[a-z0-9.-]*$', self.DOMAIN, re.IGNORECASE):
            error_exit(f"{msg_prefix} Domain name '{self.DOMAIN}' is invalid.")

        # partly check of domain name validity
        if not self.DOMAIN.lower()[0].isalpha() or \
                not self.DOMAIN.lower()[-1].isalpha():
            error_exit(f"{msg_prefix} Domain name '{self.DOMAIN}' is invalid.")

        # check domain search list domains
        for d in self.DOMAIN_SEARCH_LIST:
            # partly check of domain name validity
            if not re.match('^[a-z0-9.-]*$', d, re.IGNORECASE):
                error_exit(f"{msg_prefix} Domain search list domain name '{d}' is invalid.")

            # partly check of domain name validity
            if not d.lower()[0].isalpha() or \
                    not d.lower()[-1].isalpha():
                error_exit(f"{msg_prefix} Domain search list domain name '{d}' is invalid.")

        # check if valid lifetime is a number
        if not self.VALID_LIFETIME.isdigit():
            error_exit(f"{msg_prefix} Valid lifetime '{self.VALID_LIFETIME}' is invalid.")

        # check if preferred lifetime is a number
        if not self.PREFERRED_LIFETIME.isdigit():
            error_exit(f"{msg_prefix} Preferred lifetime '{self.PREFERRED_LIFETIME}' is invalid.")

        # check if valid lifetime is longer than preferred lifetime
        if not int(self.VALID_LIFETIME) > int(self.PREFERRED_LIFETIME):
            error_exit(f"{msg_prefix} Valid lifetime '{self.VALID_LIFETIME}' is shorter "
                       f"than preferred lifetime '{self.PREFERRED_LIFETIME}' and thus invalid.")

        # check if T1 is a number
        if not self.T1.isdigit():
            error_exit(f"{msg_prefix} T1 '{self.T1}' is invalid.")

        # check if T2 is a number
        if not self.T2.isdigit():
            error_exit(f"{msg_prefix} T2 '{self.T2}' is invalid.")

        # check T2 is not smaller than T1
        if not int(self.T2) >= int(self.T1):
            error_exit(f"{msg_prefix} T2 '{self.T2}' is shorter than T1 '{self.T1}' and thus invalid.")

        # check if T1 <= T2 <= PREFERRED_LIFETIME <= VALID_LIFETIME
        if not (int(self.T1) <= int(self.T2) <= int(self.PREFERRED_LIFETIME) <= int(self.VALID_LIFETIME)):
            error_exit(f"{msg_prefix} Time intervals T1 '{self.T1}' <= T2 '{self.T2}' <= "
                       f"preferred_lifetime '{self.PREFERRED_LIFETIME}' <= "
                       f"valid_lifetime '{self.VALID_LIFETIME}' are wrong.")

        # check server preference
        if not self.SERVER_PREFERENCE.isdigit():
            error_exit(f"{msg_prefix} Server preference '{self.SERVER_PREFERENCE}' is invalid.")
        elif not 0 <= int(self.SERVER_PREFERENCE) <= 255:
            error_exit(f"Server preference '{self.SERVER_PREFERENCE}' is invalid")

        # check information refresh time
        if not self.INFORMATION_REFRESH_TIME.isdigit():
            error_exit(f"{msg_prefix} Information refresh time '{self.INFORMATION_REFRESH_TIME}' is invalid.")
        elif not 0 < int(self.INFORMATION_REFRESH_TIME):
            error_exit(f"{msg_prefix} Information refresh time preference "
                       f"'{self.INFORMATION_REFRESH_TIME}' is pretty short.")

        # check validity of configuration source
        if self.STORE_CONFIG not in ['mysql', 'postgresql', 'sqlite', 'file', False]:
            error_exit(f"{msg_prefix} Unknown config storage type '{self.STORAGE}' is invalid.")

        # check which type of storage to use for leases
        if self.STORE_VOLATILE not in ['mysql', 'postgresql', 'sqlite']:
            error_exit(f"{msg_prefix} Unknown volatile storage type '{self.VOLATILE}' is invalid.")

        # check if database for config and volatile is equal - if any
        if self.STORE_CONFIG in ['mysql', 'postgresql'] and self.STORE_VOLATILE in ['mysql', 'postgresql']:
            if self.STORE_CONFIG != self.STORE_VOLATILE:
                error_exit(f"{msg_prefix} Storage types for database access have to be equal - "
                           f"'{self.STORE_CONFIG}' != '{self.STORE_VOLATILE}'.")

        # check validity of config file
        if self.STORE_CONFIG == 'file':
            if os.path.exists(self.STORE_FILE_CONFIG):
                if not (os.path.isfile(self.STORE_FILE_CONFIG) or
                        os.path.islink(self.STORE_FILE_CONFIG)):
                    error_exit(f"{msg_prefix} Config file '{self.STORE_FILE_CONFIG}' is no file or link.")
            else:
                error_exit(f"{msg_prefix} Config file '{self.STORE_FILE_CONFIG}' does not exist.")

        # check validity of config db sqlite file
        if self.STORE_CONFIG == 'sqlite':
            if os.path.exists(self.STORE_SQLITE_CONFIG):
                if not (os.path.isfile(self.STORE_SQLITE_CONFIG) or
                        os.path.islink(self.STORE_SQLITE_CONFIG)):
                    error_exit(f"{msg_prefix} SQLite file '{self.STORE_SQLITE_CONFIG}' is no file or link.")
            else:
                error_exit(f"{msg_prefix} SQLite file '{self.STORE_SQLITE_CONFIG}' does not exist.")

        # check validity of volatile db sqlite file
        if self.STORE_VOLATILE == 'sqlite':
            if os.path.exists(self.STORE_SQLITE_VOLATILE):
                if not (os.path.isfile(self.STORE_SQLITE_VOLATILE) or
                        os.path.islink(self.STORE_SQLITE_VOLATILE)):
                    error_exit(f"{msg_prefix} SQLite file '{self.STORE_SQLITE_VOLATILE}' is no file or link.")
            else:
                error_exit(f"{msg_prefix} SQLite file '{self.STORE_SQLITE_VOLATILE}' does not exist.")

        # check log validity
        if self.LOG:
            if self.LOG_FILE != '':
                if os.path.exists(self.LOG_FILE):
                    if not (os.path.isfile(self.LOG_FILE) or
                            os.path.islink(self.LOG_FILE)):
                        error_exit(f"{msg_prefix} Logfile '{self.LOG_FILE}' is no file or link.")
                else:
                    error_exit(f"{msg_prefix} Logfile '{self.LOG_FILE}' does not exist.")
                # check ownership of logfile
                stat_result = os.stat(self.LOG_FILE)
                if not stat_result.st_uid == pwd.getpwnam(self.USER).pw_uid:
                    error_exit(f"{msg_prefix} User {self.USER} is not owner of logfile '{self.LOG_FILE}'.")
                if not stat_result.st_gid == grp.getgrnam(self.GROUP).gr_gid:
                    error_exit(f"{msg_prefix} Group {self.GROUP} is not owner of logfile '{self.LOG_FILE}'.")

            if self.LOG_LEVEL not in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
                error_exit(f"Log level '{self.LOG_LEVEL}' is invalid")
            if self.LOG_SYSLOG:
                if self.LOG_SYSLOG_FACILITY not in ['KERN', 'USER', 'MAIL', 'DAEMON', 'AUTH',
                                                    'LPR', 'NEWS', 'UUCP', 'CRON', 'SYSLOG',
                                                    'LOCAL0', 'LOCAL1', 'LOCAL2', 'LOCAL3',
                                                    'LOCAL4', 'LOCAL5', 'LOCAL6', 'LOCAL7']:
                    error_exit(f"{msg_prefix} Syslog facility '{self.LOG_SYSLOG_FACILITY}' is invalid.")

                if self.LOG_SYSLOG_DESTINATION.startswith('/'):
                    stat_result = os.stat(self.LOG_SYSLOG_DESTINATION)
                    if not stat.S_ISSOCK(stat_result.st_mode):
                        error_exit(
                            f"{msg_prefix} Syslog destination '{self.LOG_SYSLOG_DESTINATION}' is no socket.")
                elif self.LOG_SYSLOG_DESTINATION.count(':') > 0:
                    if self.LOG_SYSLOG_DESTINATION.count(':') > 1:
                        error_exit(f"{msg_prefix} Syslog destination '{self.LOG_SYSLOG_DESTINATION}' "
                                   f"is no valid host:port destination.")

        # check authentification information
        if not self.AUTHENTICATION_INFORMATION.isalnum():
            error_exit(f"{msg_prefix} Authentification information '{self.AUTHENTICATION_INFORMATION}' "
                       f"must be alphanumeric.")

        # check validity of identification attributes
        for i in self.IDENTIFICATION:
            if i not in ['mac', 'hostname', 'duid']:
                error_exit(f"{msg_prefix} Identification must consist of 'mac', 'hostname' and/or 'duid'.")

        # check validity of identification mode
        if self.IDENTIFICATION_MODE.strip() not in ['match_all', 'match_some']:
            error_exit(f"{msg_prefix} Identification mode must be one of 'match_all' or 'match_some'.")

        # check if request rate limit seconds are a number
        if not self.REQUEST_LIMIT_TIME.isdigit():
            error_exit(f"{msg_prefix} Request limit time '{self.REQUEST_LIMIT_TIME}' is invalid.")

        # check if request rate limit count is a number
        if not self.REQUEST_LIMIT_COUNT.isdigit():
            error_exit(f"{msg_prefix} Request limit count '{self.REQUEST_LIMIT_COUNT}' is invalid.")

        # check if request rate limit blacklist release time seconds are a number
        if not self.REQUEST_LIMIT_RELEASE_TIME.isdigit():
            error_exit(f"{msg_prefix} Request limit blacklist release time "
                       f"'{self.REQUEST_LIMIT_RELEASE_TIME}' is invalid.")

        # check validity of identification attributes
        if self.REQUEST_LIMIT_IDENTIFICATION not in ['mac', 'llip']:
            error_exit(f"{msg_prefix} Request limit identification must be one of 'mac' or 'llip'.")

        # Make integers of number strings to avoid later repeated conversion
        # more to come...
        self.REQUEST_LIMIT_TIME = int(self.REQUEST_LIMIT_TIME)
        self.REQUEST_LIMIT_COUNT = int(self.REQUEST_LIMIT_COUNT)
        self.REQUEST_LIMIT_RELEASE_TIME = int(self.REQUEST_LIMIT_RELEASE_TIME)

        # cruise through classes
        # more checks to come...
        for c in self.CLASSES:
            msg_prefix = f"Class '{c}':"
            if self.CLASSES[c].ANSWER not in ['normal',
                                              'noaddress',
                                              'none']:
                error_exit(f"{msg_prefix} answer type must be one of 'normal', 'noaddress' and 'none'.")

            # check interface
            if not self.IGNORE_INTERFACE:
                for i in self.CLASSES[c].INTERFACE:
                    # also accept Linux VLAN and other definitions but interface must exist
                    if i not in get_interfaces() or not re.match('^[a-z0-9_:.%-]*$', i, re.IGNORECASE):
                        error_exit(f"{msg_prefix} Interface '{i}' is invalid.")

            # check advertised IA types
            for i in self.CLASSES[c].ADVERTISE:
                if i not in ['addresses', 'prefixes']:
                    error_exit("Only 'addresses' and 'prefixes' can be advertised.")

            # check nameserver to be given to client
            for nameserver in self.CLASSES[c].NAMESERVER:
                try:
                    decompress_ip6(nameserver)
                except Exception as err:
                    error_exit(f"{msg_prefix} Name server address '{err}' is invalid.")

            # split NTP server types into possible 3 (address, multicast, FQDN)
            # more details about this madness are available at https://tools.ietf.org/html/rfc5908
            for ntp_server in self.CLASSES[c].NTP_SERVER:
                try:
                    decompress_ip6(ntp_server)
                    # if decompressing worked it must be an address
                    if ntp_server.lower().startswith('ff'):
                        self.CLASSES[c].NTP_SERVER_DICT['MC'].append(ntp_server.lower())
                    else:
                        self.CLASSES[c].NTP_SERVER_DICT['SRV'].append(ntp_server.lower())
                except Exception as err:
                    if re.match('^[a-z0-9.-]*$', ntp_server, re.IGNORECASE):
                        self.CLASSES[c].NTP_SERVER_DICT['FQDN'].append(ntp_server.lower())
                    else:
                        error_exit(f"{msg_prefix} NTP server address '{ntp_server}' is invalid.")

            # check if T1 is a number
            if not self.CLASSES[c].T1.isdigit():
                error_exit(f"{msg_prefix} T1 '{self.CLASSES[c].T1}' is invalid.")

            # check if T2 is a number
            if not self.CLASSES[c].T2.isdigit():
                error_exit(f"{msg_prefix} T2 '{self.CLASSES[c].T2}' is invalid.")

            # check T2 is not smaller than T1
            if not int(self.CLASSES[c].T2) >= int(self.CLASSES[c].T1):
                error_exit(f"{msg_prefix} T2 '{self.CLASSES[c].T2}' is shorter "
                           f"than T1 '{self.CLASSES[c].T1}' and thus invalid.")

            # check every single address of a class
            for a in self.CLASSES[c].ADDRESSES:
                msg_prefix = f"Class '{c}' Address type '{a}':"
                # test if used addresses are defined
                if a not in self.ADDRESSES:
                    error_exit(f"{msg_prefix} Address type '{a}' is not defined.")

                # test validity of category
                if self.ADDRESSES[a].CATEGORY.strip() not in ['eui64', 'fixed', 'range', 'random', 'mac', 'id', 'dns']:
                    error_exit(f"{msg_prefix} Category '{self.ADDRESSES[a].CATEGORY}' is invalid. "
                               f"Category must be one of 'eui64', 'fixed', 'range', 'random', 'mac', 'id' and 'dns'.")

                # test validity of pattern - has its own error output
                self.ADDRESSES[a].build_prototype()
                # test existence of category specific variable in pattern
                if self.ADDRESSES[a].CATEGORY == 'range':
                    if not re.match('^[0-9a-f]{1,4}-[0-9a-f]{1,4}$', self.ADDRESSES[a].RANGE, re.IGNORECASE):
                        error_exit(f"{msg_prefix} Range '{self.ADDRESSES[a].RANGE}' is not valid.")
                    if not 0 < self.ADDRESSES[a].PATTERN.count('$range$') < 2:
                        error_exit(f"{msg_prefix} Pattern '{self.ADDRESSES[a].PATTERN.strip()}' contains wrong "
                                   f"number of '$range$' variables for category 'range'.")
                    elif not self.ADDRESSES[a].PATTERN.endswith('$range$'):
                        error_exit(f"{msg_prefix} Pattern '{self.ADDRESSES[a].PATTERN.strip()}' must end "
                                   f"with '$range$' variable for category 'range'.")

                if self.ADDRESSES[a].CATEGORY == 'mac':
                    if not 0 < self.ADDRESSES[a].PATTERN.count('$mac$') < 2:
                        error_exit(f"{msg_prefix} Pattern '{self.ADDRESSES[a].PATTERN.strip()}' contains wrong "
                                   f"number of '$mac$' variables for category 'mac'.")

                if self.ADDRESSES[a].CATEGORY == 'id':
                    if not self.ADDRESSES[a].PATTERN.count('$id$') == 1:
                        error_exit(f"{msg_prefix} Pattern '{self.ADDRESSES[a].PATTERN.strip()}' contains wrong "
                                   f"number of '$id$' variables for category 'id'.")

                if self.ADDRESSES[a].CATEGORY == 'random':
                    if not self.ADDRESSES[a].PATTERN.count('$random64$') == 1:
                        error_exit(f"{msg_prefix} Pattern '{self.ADDRESSES[a].PATTERN.strip()}' contains wrong "
                                   f"number of '$random64$' variables for category 'random'.")

                if self.ADDRESSES[a].CATEGORY == 'dns':
                    if not len(self.NAMESERVER) > 0:
                        error_exit("Address of category 'dns' needs a set nameserver.")

                # check ia_type
                if not self.ADDRESSES[a].IA_TYPE.strip().lower() in ['na', 'ta']:
                    error_exit(f"{msg_prefix}: IA type '{self.ADDRESSES[a].IA_TYPE.strip()}' "
                               f"must be one of 'na' or 'ta'.")

                # check if valid lifetime is a number
                if not self.ADDRESSES[a].VALID_LIFETIME.isdigit():
                    error_exit(f"{msg_prefix} Valid lifetime '{self.ADDRESSES[a].VALID_LIFETIME}' is invalid.")

                # check if preferred lifetime is a number
                if not self.ADDRESSES[a].PREFERRED_LIFETIME.isdigit():
                    error_exit(
                        f"{msg_prefix} Preferred lifetime '{self.ADDRESSES[a].PREFERRED_LIFETIME}' is invalid.")

                # check if valid lifetime is longer than preferred lifetime
                if not int(self.ADDRESSES[a].VALID_LIFETIME) >= int(self.ADDRESSES[a].PREFERRED_LIFETIME):
                    error_exit(f"{msg_prefix} Valid lifetime '{self.ADDRESSES[a].VALID_LIFETIME}' is shorter "
                               f"than preferred lifetime '{self.ADDRESSES[a].PREFERRED_LIFETIME}' and thus invalid.")

                # check if T1 <= T2 <= PREFERRED_LIFETIME <= VALID_LIFETIME
                if not (int(self.CLASSES[c].T1) <= int(self.CLASSES[c].T2) <=
                        int(self.ADDRESSES[a].PREFERRED_LIFETIME) <= int(self.ADDRESSES[a].VALID_LIFETIME)):
                    error_exit(f"{msg_prefix} Time intervals T1 '{self.CLASSES[c].T1}' <= "
                               f"T2 '{self.CLASSES[c].T2}' <= "
                               f"preferred_lifetime '{self.ADDRESSES[a].PREFERRED_LIFETIME}' <= "
                               f"valid_lifetime '{self.ADDRESSES[a].VALID_LIFETIME}' are wrong.")

            # check every single bootfile of a class
            for b in self.CLASSES[c].BOOTFILES:
                msg_prefix = f"Bootfile '{c}' BOOTFILE type '{b}':"
                # test if used bootfiles are defined
                if b not in self.BOOTFILES:
                    error_exit(f"{msg_prefix} Bootfile type '{b}' is not defined.")

            # check every single prefix of a class
            for p in self.CLASSES[c].PREFIXES:
                msg_prefix = f"Class '{c}' PREFIX type '{p}':"
                # test if used addresses are defined
                if p not in self.PREFIXES:
                    error_exit(f"{msg_prefix} Prefix type '{p}' is not defined.")

                # test validity of category
                if self.PREFIXES[p].CATEGORY.strip() not in ['range', 'id']:
                    error_exit(f"{msg_prefix} Category 'self.PREFIXES[p].CATEGORY' is invalid. "
                               f"Category must be 'range' or 'id'.")

                # test validity of pattern - has its own error output
                self.PREFIXES[p].build_prototype()
                # test existence of category specific variable in pattern
                if self.PREFIXES[p].CATEGORY == 'range':
                    if not re.match('^[0-9a-f]{1,4}-[0-9a-f]{1,4}$', self.PREFIXES[p].RANGE, re.IGNORECASE):
                        error_exit(f"{msg_prefix} Range '{self.PREFIXES[p].RANGE}' is not valid.")
                    if not 0 < self.PREFIXES[p].PATTERN.count('$range$') < 2:
                        error_exit(f"{msg_prefix} Pattern '{self.PREFIXES[p].PATTERN.strip()}' contains wrong "
                                   f"number of '$range$' variables for category 'range'.")
                    elif self.PREFIXES[p].PATTERN.endswith('$range$'):
                        error_exit(f"{msg_prefix} Pattern '{self.PREFIXES[p].PATTERN.strip()}' must not end "
                                   f"with '$range$' variable for category 'range'.")

                # check if valid lifetime is a number
                if not self.PREFIXES[p].VALID_LIFETIME.isdigit():
                    error_exit(f"{msg_prefix} Valid lifetime '{self.PREFIXES[p].VALID_LIFETIME}' is invalid.")

                # check if preferred lifetime is a number
                if not self.PREFIXES[p].PREFERRED_LIFETIME.isdigit():
                    error_exit(
                        f"{msg_prefix} Preferred lifetime '{self.PREFIXES[p].PREFERRED_LIFETIME}' is invalid.")

                # check if valid lifetime is longer than preferred lifetime
                if not int(self.PREFIXES[p].VALID_LIFETIME) >= int(self.PREFIXES[p].PREFERRED_LIFETIME):
                    error_exit(f"{msg_prefix} Valid lifetime '{self.PREFIXES[p].VALID_LIFETIME}' is shorter "
                               f"than preferred lifetime '{self.PREFIXES[p].PREFERRED_LIFETIME}' and thus invalid.")

                # check if T1 <= T2 <= PREFERRED_LIFETIME <= VALID_LIFETIME
                if not (int(self.CLASSES[c].T1) <= int(self.CLASSES[c].T2) <=
                        int(self.PREFIXES[p].PREFERRED_LIFETIME) <= int(self.PREFIXES[p].VALID_LIFETIME)):
                    error_exit(f"{msg_prefix} Time intervals T1 '{self.CLASSES[c].T1}' <= "
                               f"T2 '{self.CLASSES[c].T2}' <= "
                               f"preferred_lifetime '{self.PREFIXES[p].PREFERRED_LIFETIME}' <= "
                               f"valid_lifetime '{self.PREFIXES[p].VALID_LIFETIME}' are wrong.")

                # check if prefix is a valid number
                if not self.PREFIXES[p].LENGTH.isdigit():
                    error_exit(f"{msg_prefix} Prefix length '{self.PREFIXES[p].LENGTH}' is invalid.")
                if not 0 <= int(self.PREFIXES[p].LENGTH) <= 128:
                    error_exit(f"{msg_prefix} Prefix length '{self.PREFIXES[p].LENGTH}' must be in range 0-128.")

        # cruise through bootfiles
        # more checks to come...
        for b in self.BOOTFILES:
            msg_prefix = f"Bootfile '{b}':"
            bootfile_url = self.BOOTFILES[b].BOOTFILE_URL

            if bootfile_url is None or bootfile_url == '':
                error_exit(f"{msg_prefix} Bootfile url parameter must be set and is not allowed to be empty.")


class ConfigObject:
    """
        class providing methods both for addresses and prefixes
    """

    def build_prototype(self, pattern=None):
        """
            build prototype of pattern for later comparison with leases
        """

        # if called with de-$prefix$-ed pattern use it
        if pattern is None:
            prototype = self.PATTERN
        else:
            prototype = pattern

        # inject prefix later so jump out here now
        if '$prefix$' in prototype:
            self.PROTOTYPE = prototype
            return

        # check different client address categories - to be extended!
        if self.CATEGORY in ['mac', 'id', 'range', 'random']:
            if self.CATEGORY == 'mac':
                prototype = prototype.replace('$mac$', 'xxxx:xxxx:xxxx')
            elif self.CATEGORY == 'id':
                prototype = prototype.replace('$id$', 'xxxx')
            elif self.CATEGORY == 'random':
                prototype = prototype.replace('$random64$', 'xxxx:xxxx:xxxx:xxxx')
            elif self.CATEGORY == 'range':
                prototype = prototype.replace('$range$', 'xxxx')
            try:
                # build complete 'address' and ignore all the Xs (strict=False)
                # all X will become x
                prototype = decompress_ip6(prototype, strict=False)
            except Exception as err:
                error_exit(f"Address type '{self.TYPE}' address pattern '{self.PATTERN}' is not valid: {err}")

        self.PROTOTYPE = prototype

    def inject_dynamic_prefix_into_prototype(self, dynamic_prefix):
        """
            called from main to put then known dynamic prefix into protoype
        """
        if '$prefix$' in self.PATTERN:
            prefix_pattern = self.PATTERN.replace('$prefix$', dynamic_prefix)
            self.build_prototype(prefix_pattern)

    def matches_prototype(self, address):
        """
            test if given address matches prototype and therefore this address' DNS zone
            information might be used
            only used for address types, not client instances
        """
        match = False
        # compare all chars of address and prototype, if they do match or
        # prototype has placeholder X return finally True, otherwise stop
        # at the first difference and give back False
        for i in range(32):
            if self.PROTOTYPE[i] == address[i] or self.PROTOTYPE[i] == 'x':
                match = True
            else:
                match = False
                break
        return match


class Address(ConfigObject):
    """
        class for address definition, used for config
    """

    def __init__(self,
                 address=None,
                 ia_type='na',
                 category='random',
                 pattern='2001:db8::$random64$',
                 preferred_lifetime=0,
                 valid_lifetime=0,
                 atype='default',
                 aclass='default',
                 prototype='',
                 arange='',
                 dns_update=False,
                 dns_zone='',
                 dns_rev_zone='8.b.d.0.1.0.0.2.ip6.arpa',
                 dns_ttl='0',
                 valid=True):
        self.CATEGORY = category
        self.PATTERN = pattern
        self.IA_TYPE = ia_type
        self.PREFERRED_LIFETIME = preferred_lifetime
        self.VALID_LIFETIME = valid_lifetime
        self.ADDRESS = address
        self.RANGE = arange.lower()
        # because 'class' is a python keyword we use 'client_class' here
        self.CLASS = aclass
        # same with type
        self.TYPE = atype
        # a prototypical address to be compared with leases given by
        # clients - if prototype and lease address kind of match
        # give back the lease as valid
        self.PROTOTYPE = prototype
        # flag for updating address in DNS or not
        self.DNS_UPDATE = dns_update
        # DNS zone data
        self.DNS_ZONE = dns_zone.lower()
        self.DNS_REV_ZONE = dns_rev_zone.lower()
        self.DNS_TTL = dns_ttl
        # flag invalid addresses as invalid, valid ones as valid
        self.VALID = valid


class Prefix(ConfigObject):
    """
        class for delegated prefix definition
    """

    def __init__(self,
                 prefix=None,
                 pattern='2001:db8:$range$::',
                 prange='1000-1fff',
                 category='range',
                 length='48',
                 preferred_lifetime=0,
                 valid_lifetime=0,
                 ptype='default',
                 pclass='default',
                 valid=True,
                 route_link_local=False):
        self.PREFIX = prefix
        self.PATTERN = pattern
        self.RANGE = prange.lower()
        self.CATEGORY = category
        self.LENGTH = length
        self.PREFERRED_LIFETIME = preferred_lifetime
        self.VALID_LIFETIME = valid_lifetime
        self.TYPE = ptype
        self.CLASS = pclass
        self.VALID = valid
        self.ROUTE_LINK_LOCAL = route_link_local


class Class:
    """
        class for class definition
    """

    def __init__(self, name=''):
        self.NAME = name
        self.ADDRESSES = list()
        self.PREFIXES = list()
        self.BOOTFILES = list()
        self.NAMESERVER = ''
        self.NTP_SERVER = ''
        # Auxiliary options, derived from self.NTP_SERVER
        self.NTP_SERVER_DICT = {'SRV': [], 'MC': [], 'FQDN': []}
        self.FILTER_MAC = ''
        self.FILTER_HOSTNAME = ''
        self.FILTER_DUID = ''
        self.IDENTIFICATION_MODE = 'match_all'
        # RENEW time
        self.T1 = 0
        # REBIND time
        self.T2 = 0
        # at which interface this class of clients is served
        self.INTERFACE = list()
        # in certain cases it might be useful not to give any address to clients, for example if only a defined group
        # of hosts should get IPv6 addresses and others not. They will get a 'NoAddrsAvail' handler if this option
        # is set to 'noaddress' or no answer at all if set to 'none'
        self.ANSWER = 'normal'
        # which IA_* should this class supply - addresses, prefixes or both?
        # shouldn't be an empty list because in this case the class would not make sense at all
        # as default only addresses will be advertised
        self.ADVERTISE = ['addresses']
        # commands or scripts to be called for setting and removing routes for prefixes
        self.CALL_UP = ''
        self.CALL_DOWN = ''


class BootFile:
    """
        class for netboot defintion
    """

    def __init__(self, name=''):
        self.NAME = name
        # PXE client architecture (Option 61)
        self.CLIENT_ARCHITECTURE = ''
        # PXE bootfile URL (Option 59)
        self.BOOTFILE_URL = ''
        # User class (Option 15)
        self.USER_CLASS = ''


def generate_duid():
    """
    Creates a DUID for the server - needed if none exists or is given
    :return:
    """
    return f'00010001{int(time.time()):08x}{uuid.getnode():012x}'


# singleton-like central instance
cfg = Config()



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""Module dhcpy6d"""

import socket
import socketserver
import struct

from .config import cfg
from .globals import (collected_macs,
                      IF_NAME,
                      IF_NUMBER,
                      NC,
                      OS,
                      timer)
from .helpers import (colonify_ip6,
                      colonify_mac,
                      correct_mac,
                      decompress_ip6,
                      NeighborCacheRecord)
from .log import log
from .storage import volatile_store


class UDPMulticastIPv6(socketserver.UnixDatagramServer):
    """
        modify server_bind to work with multicast
        add DHCPv6 multicast group ff02::1:2
    """
    def server_bind(self):
        """
            multicast & python: http://code.activestate.com/recipes/442490/
        """
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # multicast parameters
        # hop is one because it is all about the same subnet
        self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, 0)
        self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, 1)

        for i in cfg.INTERFACE:
            IF_NAME[i] = socket.if_nametoindex(i)
            IF_NUMBER[IF_NAME[i]] = i
            if_number = struct.pack('I', IF_NAME[i])
            #mgroup = socket.inet_pton(socket.AF_INET6, cfg.MCAST) + if_number
            # no need vor variable.... the DHCPv6 multicast address is predefined
            mgroup = socket.inet_pton(socket.AF_INET6, 'ff02::1:2') + if_number

            # join multicast group - should work definitively if not ignoring interface at startup
            if cfg.IGNORE_INTERFACE:
                try:
                    self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, mgroup)
                except Exception as err:
                    print(err)
            else:
                self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, mgroup)

        # bind socket to server address
        self.socket.bind(self.server_address)

        # attempt to avoid blocking
        self.socket.setblocking(False)

        # some more requests?
        self.request_queue_size = 100



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import sys
import traceback

from ..config import cfg
from ..helpers import error_exit

from .store import DB


class DBMySQL(DB):
    """
    access MySQL and MariaDB
    """
    QUERY_TABLES = f"SELECT table_name FROM information_schema.tables WHERE table_schema = '{cfg.STORE_DB_DB}'"

    def db_connect(self):
        """
            Connect to database server according to database type
        """
        # try to get some MySQL/MariaDB-module imported
        try:
            if 'MySQLdb' not in sys.modules:
                import MySQLdb
            self.db_module = sys.modules['MySQLdb']
        except:
            try:
                if 'pymsql' not in sys.modules:
                    import pymysql
                self.db_module = sys.modules['pymysql']
            except:
                error_exit('ERROR: Cannot find module MySQLdb or PyMySQL. Please install one of them to proceed.')
        try:
            self.connection = self.db_module.connect(host=cfg.STORE_DB_HOST,
                                                     db=cfg.STORE_DB_DB,
                                                     user=cfg.STORE_DB_USER,
                                                     passwd=cfg.STORE_DB_PASSWORD)
            self.connection.autocommit(True)
            self.cursor = self.connection.cursor()
            self.connected = True
        except:
            traceback.print_exc(file=sys.stdout)
            sys.stdout.flush()
            self.connected = False

        return self.connected

    def db_query(self, query):
        """
        execute DB query
        """
        try:
            self.cursor.execute(query)
        except self.db_module.IntegrityError:
            return 'INSERT_ERROR'
        except Exception as err:
            # try to reestablish database connection
            print(f'Error: {str(err)}')
            print(f'Query: {query}')
            if not self.db_connect():
                return None
            else:
                try:
                    self.cursor.execute(query)
                except:
                    traceback.print_exc(file=sys.stdout)
                    sys.stdout.flush()
                    self.connected = False
                    return None

        result = self.cursor.fetchall()
        return result



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from .config import cfg
from .globals import (route_queue,
                      timer)
from .log import log
from .storage import volatile_store


class Route:
    """
        store data of a route which should be given to an external application
        router is here the prefix requesting host
    """

    def __init__(self, prefix, length, router):
        self.prefix = prefix
        self.length = length
        self.router = router


def modify_route(transaction, mode):
    """
        called when route has to be set - calls itself any external script or something like that
    """
    # check if client is already set - otherwise crashes
    if transaction.client is not None:
        # only do anything if class of client has something defined to be called
        if (mode == 'up' and cfg.CLASSES[transaction.client.client_class].CALL_UP != '') or \
                (mode == 'down' and cfg.CLASSES[transaction.client.client_class].CALL_DOWN != ''):
            # collect possible prefixes, lengths and router ip addresses in list
            routes = list()
            for prefix in transaction.client.prefixes:
                # use LinkLocal Address of client if wanted
                if prefix.ROUTE_LINK_LOCAL:
                    router = transaction.client_llip
                else:
                    if len(transaction.client.addresses) == 1:
                        router = transaction.client.addresses[0].ADDRESS
                    else:
                        router = None
                        log.error(
                            'modify_route: client needs exactly 1 address to be used as router to delegated prefix')
                if router is not None:
                    routes.append(Route(prefix.PREFIX, prefix.LENGTH, router))

            if mode == 'up':
                call = cfg.CLASSES[transaction.client.client_class].CALL_UP
            elif mode == 'down':
                call = cfg.CLASSES[transaction.client.client_class].CALL_DOWN
            else:
                # should not happen but just in case
                call = ''

            # call executables here
            for route in routes:
                route_queue.put((mode, call, route.prefix, route.length, route.router))


def manage_prefixes_routes():
    """
        delete or add inactive or active routes according to the prefixes in database
    """
    volatile_store.release_free_prefixes(timer.time)
    inactive_prefixes = volatile_store.get_inactive_prefixes()
    active_prefixes = volatile_store.get_active_prefixes()

    for prefix in inactive_prefixes:
        length, router, pclass = volatile_store.get_route(prefix)
        if pclass in cfg.CLASSES:
            route_queue.put(('down', cfg.CLASSES[pclass].CALL_DOWN, prefix, length, router))

    for prefix in active_prefixes:
        length, router, pclass = volatile_store.get_route(prefix)
        if pclass in cfg.CLASSES:
            route_queue.put(('up', cfg.CLASSES[pclass].CALL_UP, prefix, length, router))



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import sys
import traceback

from ..config import cfg
from ..globals import collected_macs
from ..helpers import (decompress_ip6,
                       error_exit,
                       listify_option,
                       NeighborCacheRecord,
                       convert_prefix_inline)
from .schemas import (legacy_adjustments,
                      MYSQL_SQLITE)


class ClientConfig:
    """
        static client settings object to be stuffed into Hosts dict of Textfile store
    """
    def __init__(self, hostname='', client_class='default', duid='', address=None, prefix=None, mac=None, host_id=''):
        self.HOSTNAME = hostname
        # MACs
        if type(mac) == list:
            self.MAC = mac
        else:
            self.MAC = listify_option(mac)
        # fixed addresses
        if address:
            self.ADDRESS = list()
            if type(address) == list:
                addresses = address
            else:
                addresses = listify_option(address)
            for a in addresses:
                self.ADDRESS.append(decompress_ip6(a))
        else:
            self.ADDRESS = None

        # fixed prefix
        if prefix:
            self.PREFIX = list()
            if type(prefix) == list:
                prefixes = prefix
            else:
                prefixes = listify_option(prefix)
            for p in prefixes:
                self.PREFIX.append(convert_prefix_inline(p))
        else:
            self.PREFIX = None

        self.CLASS = client_class
        self.ID = host_id
        self.DUID = duid


class ClientConfigDicts:
    """
        class for storing client config snippet from DB - used in SQLite and MySQL Storage
    """

    def __init__(self):
        self.hosts = {}
        self.index_mac = {}
        self.index_duid = {}


class Store:
    """
    abstract class to present MySQL or SQLite or Postgresql
    """
    # put SQL schemas here to be in reach of all storage types
    SCHEMAS = MYSQL_SQLITE

    # query to get tables - different in every SQL storage
    # if no tables exist they will be created by create_tables()
    QUERY_TABLES = ''

    # increasing number of SQL schema versions
    version = 3

    # link to used database module
    db_module = None

    # flag for config database prefix support
    config_prefix_support = False

    def __init__(self, query_queue, answer_queue):
        self.query_queue = query_queue
        self.answer_queue = answer_queue
        # table names used for database storage - MySQL additionally needs the database name
        self.table_leases = 'leases'
        self.table_prefixes = 'prefixes'
        self.table_macs_llips = 'macs_llips'
        self.table_hosts = 'hosts'
        self.table_routes = 'routes'
        # flag to check if connection is OK
        self.connected = False
        # storage of query answers
        self.answers = {}

    def query(self, query):
        """
        put queries received into query queue and return the answers from answer queue
        """
        if query in self.answers:
            answer = self.answers.pop(query)
        else:
            answer = None
            while answer is None:
                self.query_queue.put(query)
                self.answers.update(self.answer_queue.get())
                # just make sure the right answer comes back
                if query in self.answers:
                    answer = self.answers.pop(query)
        return answer

    def clean_query_answer(method):
        """
        decorate repeatedly but not everywhere used cleaning of query answer
        """

        def decoration_function(self, *args, **kwargs):
            # run decorated method
            answer = method(self, *args, **kwargs)
            # clean answer
            # SQLite returns list, MySQL tuple - in case someone wonders here...
            if not (answer == [] or answer == () or answer is None):
                return answer[0][0]
            else:
                return None

        return decoration_function

    @clean_query_answer
    def get_db_version(self):
        """
        return stored version if dhcpy6d DB

        """
        try:
            # will be cleaned by decorator so default answer value is a little bit unintuitive
            answer = [['0']]
            if 'meta' in self.get_tables():
                answer = self.query("SELECT item_value FROM meta WHERE item_key = 'version'")
            return answer
        except:
            # no table 'meta' and no key 'version'
            return answer

    def get_tables(self):
        """
        return tables - no turntables
        """
        tables = []
        # every DB type needs another query for tables
        query = self.QUERY_TABLES
        answer = self.query(query)
        if len(answer) > 0:
            tables = [x[0] for x in answer]
        return tables

    def create_tables(self):
        """
        create tables in different storage types - information about schemas comes from schemas.py
        """
        for table in self.SCHEMAS:
            query = self.SCHEMAS[table]
            self.cursor.execute(query)
        # set initial version
        self.cursor.execute(f"INSERT INTO meta (item_key, item_value) VALUES ('version', '{self.version}')")

    def check_storage(self):
        """
        check if all databases/storage is ready and up to date
        """
        tables = self.get_tables()
        # if there are no tables they have to be created
        if len(tables) == 0:
            self.create_tables()
        # otherwise they migth need just some updates
        else:
            legacy_adjustments(self)

    def check_config_prefixes_support(self):
        """
        check if client config database contains prefixes - if not just do not ask for prefixes in the future
        """
        # has to be checked only for databases
        if cfg.STORE_CONFIG and cfg.STORE_CONFIG != 'file':
            # first check if database has config information at all
            try:
                self.cursor.execute(f"SELECT hostname, mac, duid, class, address, id FROM {self.table_hosts} LIMIT 1")
            except Exception as error:
                error_exit(f"Config database has problem: {error.args[-1]}")
            # if config information exists check if it also has prefixes
            try:
                self.cursor.execute(f"SELECT hostname, mac, duid, class, address, prefix, id FROM {self.table_hosts} LIMIT 1")
            except:
                return False
            # when query was ok prefix support exists
            self.config_prefix_support = True
            return True

    def store(self, transaction, now):
        """
        store lease in lease DB
        """
        # only if client exists
        if transaction.client:
            for a in transaction.client.addresses:
                if a.ADDRESS is not None:
                    query = f"SELECT address FROM {self.table_leases} WHERE address = '{a.ADDRESS}'"
                    answer = self.query(query)
                    if answer is not None:
                        # if address is not leased yet add it
                        if len(answer) == 0:
                            query = f"INSERT INTO {self.table_leases} (address, active, last_message, " \
                                    f"preferred_lifetime, valid_lifetime, hostname, type, category, ia_type, " \
                                    f"class, mac, duid, iaid, last_update, preferred_until, valid_until) " \
                                    f"VALUES ('{a.ADDRESS}', " \
                                    f"'1', " \
                                    f"'{transaction.last_message_received_type}', " \
                                    f"'{a.PREFERRED_LIFETIME}', " \
                                    f"'{a.VALID_LIFETIME}', " \
                                    f"'{transaction.client.hostname}', " \
                                    f"'{a.TYPE}', " \
                                    f"'{a.CATEGORY}', " \
                                    f"'{a.IA_TYPE}', " \
                                    f"'{transaction.client.client_class}', " \
                                    f"'{transaction.mac}', " \
                                    f"'{transaction.duid}', " \
                                    f"'{transaction.iaid}', " \
                                    f"'{now}', " \
                                    f"'{now + int(a.PREFERRED_LIFETIME)}', " \
                                    f"'{now + int(a.VALID_LIFETIME)}')"
                            answer = self.query(query)
                            # for unknown reasons sometime a lease shall be inserted which already exists
                            # in this case go further (aka continue) and do an update instead of an insert
                            if answer == 'INSERT_ERROR':
                                print('IntegrityError:', query)
                            else:
                                # jump to next item of loop
                                continue
                        # otherwise update it if not a random address
                        if a.CATEGORY != 'random':
                            query = f"UPDATE {self.table_leases} " \
                                    f"SET active = 1, " \
                                    f"last_message = {transaction.last_message_received_type}, " \
                                    f"preferred_lifetime = '{a.PREFERRED_LIFETIME}', " \
                                    f"valid_lifetime = '{a.VALID_LIFETIME}', " \
                                    f"hostname = '{transaction.client.hostname}', " \
                                    f"type = '{a.TYPE}', " \
                                    f"category = '{a.CATEGORY}', " \
                                    f"ia_type = '{a.IA_TYPE}', " \
                                    f"class = '{transaction.client.client_class}', " \
                                    f"mac = '{transaction.mac}', " \
                                    f"duid = '{transaction.duid}', " \
                                    f"iaid = '{transaction.iaid}', " \
                                    f"last_update = '{now}', " \
                                    f"preferred_until = '{now + int(a.PREFERRED_LIFETIME)}', " \
                                    f"valid_until = '{now + int(a.VALID_LIFETIME)}' " \
                                    f"WHERE address = '{a.ADDRESS}'"
                        else:
                            # set last message type of random address
                            query = f"UPDATE {self.table_leases} " \
                                    f"SET active = 1, " \
                                    f"last_message = {transaction.last_message_received_type} " \
                                    f"WHERE address = '{a.ADDRESS}'"
                        self.query(query)

            for p in transaction.client.prefixes:
                if p.PREFIX is not None:
                    query = f"SELECT prefix FROM {self.table_prefixes} WHERE prefix = '{p.PREFIX}'"
                    answer = self.query(query)
                    if answer is not None:
                        # if prefix is not leased yet add it
                        if len(answer) == 0:
                            query = f"INSERT INTO {self.table_prefixes} (prefix, length, active, last_message, " \
                                    f"preferred_lifetime, valid_lifetime, hostname, type, category, class, mac, duid, " \
                                    f"iaid, last_update, preferred_until, valid_until) " \
                                    f"VALUES ('{p.PREFIX}', " \
                                    f"'{p.LENGTH}', " \
                                    f"1, " \
                                    f"'{transaction.last_message_received_type}', " \
                                    f"'{p.PREFERRED_LIFETIME}', " \
                                    f"'{p.VALID_LIFETIME}', " \
                                    f"'{transaction.client.hostname}', " \
                                    f"'{p.TYPE}', " \
                                    f"'{p.CATEGORY}', " \
                                    f"'{transaction.client.client_class}', " \
                                    f"'{transaction.mac}', " \
                                    f"'{transaction.duid}', " \
                                    f"'{transaction.iaid}', " \
                                    f"'{now}', " \
                                    f"'{now + int(p.PREFERRED_LIFETIME)}', " \
                                    f"'{now + int(p.VALID_LIFETIME)}')"
                            answer = self.query(query)
                            # for unknow reasons sometime a lease shall be inserted which already exists
                            # in this case go further (aka continue) and do an update instead of an insert
                            # doing this here for prefixes is just a precautional measure
                            if answer != 'INSERT_ERROR':
                                continue
                        # otherwise update it if not a random prefix
                        # anyway right now only the categories 'range' and 'id' exist
                        if p.CATEGORY != 'random':
                            query = f"UPDATE {self.table_prefixes} SET active = 1, " \
                                    f"last_message = {transaction.last_message_received_type}, " \
                                    f"preferred_lifetime = '{p.PREFERRED_LIFETIME}', " \
                                    f"valid_lifetime = '{p.VALID_LIFETIME}', " \
                                    f"hostname = '{transaction.client.hostname}', " \
                                    f"type = '{p.TYPE}', " \
                                    f"category = '{p.CATEGORY}', " \
                                    f"class = '{transaction.client.client_class}', " \
                                    f"mac = '{transaction.mac}', " \
                                    f"duid = '{transaction.duid}', " \
                                    f"iaid = '{transaction.iaid}', " \
                                    f"last_update = '{now}', " \
                                    f"preferred_until = '{now + int(p.PREFERRED_LIFETIME)}', " \
                                    f"valid_until = '{now + int(p.VALID_LIFETIME)}' " \
                                    f"WHERE prefix = '{p.PREFIX}'"
                        else:
                            # set last message type of random prefix
                            query = f"UPDATE {self.table_prefixes} " \
                                    f"SET last_message = {transaction.last_message_received_type}, " \
                                    f"active = 1 " \
                                    f"WHERE prefix = '{p.PREFIX}'"
                        self.query(query)
            return True
        # if no client -> False
        return False

    @clean_query_answer
    def store_route(self, prefix, length, router, now):
        """
        store route in database to keep track of routes and be able to delete them later
        """
        query = f"SELECT prefix FROM {self.table_routes} WHERE prefix = '{prefix}'"
        if self.query is not None:
            if len(self.query(query)) == 0:
                query = f"INSERT INTO {self.table_routes} VALUES ('{prefix}', {length}, '{router}', {now})"
                return self.query(query)
            elif len(self.query(query)) == 1:
                query = f"UPDATE {self.table_routes} SET prefix = '{prefix}', length = {length}, " \
                        f"router = '{router}', last_update = {now} WHERE prefix = '{prefix}'"
                return self.query(query)
            return None
        else:
            return None

    @clean_query_answer
    def get_range_lease_for_recycling(self, prefix='', range_from='', range_to='', duid='', mac=''):
        """
        ask DB for last known leases of an already known host to be recycled
        this is most useful for CONFIRM-requests that will get a not-available-answer but get an
        ADVERTISE with the last known-as-good address for a client
        SOLICIT message type is 1
        """
        query = f"SELECT address FROM {self.table_leases} WHERE " \
                f"category = 'range' AND " \
                f"'{prefix + range_from}' <= address AND " \
                f"address <= '{prefix + range_to}' AND " \
                f"duid = '{duid}' AND " \
                f"mac = '{mac}' AND " \
                f"last_message != 1 " \
                f"ORDER BY last_update DESC LIMIT 1"
        return self.query(query)

    @clean_query_answer
    def get_range_prefix_for_recycling(self, prefix='', length='', range_from='', range_to='', duid='', mac=''):
        """
        ask DB for last known prefixes of an already known host to be recycled
        this is most useful for CONFIRM-requests that will get a not-available-answer but get an
        ADVERTISE with the last known-as-good address for a client
        SOLICIT message type is 1
        """
        query = f"SELECT prefix FROM {self.table_prefixes} WHERE " \
                f"category = 'range' AND " \
                f"'{prefix + range_from + ((128 - int(length)) // 4) * '0'}' <= prefix AND " \
                f"prefix <= '{prefix + range_to + ((128 - int(length)) // 4) * '0'}' AND " \
                f"length = '{length}' AND " \
                f"duid = '{duid}' AND " \
                f"mac = '{mac}' AND " \
                f"last_message != 1 " \
                f"ORDER BY last_update DESC LIMIT 1"
        return self.query(query)

    @clean_query_answer
    def get_highest_range_lease(self, prefix='', range_from='', range_to=''):
        """
        ask DB for highest known leases - if necessary range sensitive
        """
        query = f"SELECT address FROM {self.table_leases} WHERE active = 1 AND " \
                f"category = 'range' AND " \
                f"'{prefix + range_from}' <= address and address <= '{prefix + range_to}' ORDER BY address DESC LIMIT 1"
        return self.query(query)

    @clean_query_answer
    def get_highest_range_prefix(self, prefix='', length='', range_from='', range_to=''):
        """
        ask DB for highest known prefix - if necessary range sensitive
        """
        query = f"SELECT prefix FROM {self.table_prefixes} WHERE active = 1 AND " \
                f"category = 'range' AND " \
                f"'{prefix + range_from + ((128 - int(length)) // 4) * '0'}' <= prefix AND " \
                f"prefix <= '{prefix + range_to + ((128 - int(length)) // 4) * '0'}' AND " \
                f"length = '{length}' ORDER BY prefix DESC LIMIT 1"
        return self.query(query)

    @clean_query_answer
    def get_oldest_inactive_range_lease(self, prefix='', range_from='', range_to=''):
        """
        ask DB for oldest known inactive lease to minimize chance of collisions
        ordered by valid_until to get leases that are free as long as possible
        """
        query = f"SELECT address FROM {self.table_leases} WHERE active = 0 AND " \
                f"category = 'range' AND " \
                f"'{prefix + range_from}' <= address AND " \
                f"address <= '{prefix + range_to}' ORDER BY valid_until ASC LIMIT 1"
        return self.query(query)

    @clean_query_answer
    def get_oldest_inactive_range_prefix(self, prefix='', length='', range_from='', range_to=''):
        """
        ask DB for oldest known inactive prefix to minimize chance of collisions
        ordered by valid_until to get leases that are free as long as possible
        """
        query = f"SELECT prefix FROM {self.table_prefixes} WHERE active = 0 AND " \
                f"category = 'range' AND " \
                f"'{prefix + range_from + ((128 - int(length)) // 4) * '0'}' <= prefix AND " \
                f"prefix <= '{prefix + range_to + ((128 - int(length)) // 4) * '0'}' AND " \
                f"length = '{length}' " \
                f"ORDER BY valid_until ASC LIMIT 1"
        return self.query(query)

    def get_host_lease(self, address):
        """
        get the hostname, DUID, MAC and IAID to verify a lease to delete its address in the DNS
        """
        query = f"SELECT DISTINCT hostname, duid, mac, iaid FROM {self.table_leases} WHERE address='{address}'"
        answer = self.query(query)
        if answer is not None:
            if len(answer) > 0:
                if len(answer[0]) > 0:
                    return answer[0]
                else:
                    # calling method expects quartet of hostname, duid, mac, iad - get None if nothing there
                    return None, None, None, None
            else:
                return None, None, None, None
        else:
            return None, None, None, None

    def get_active_prefixes(self):
        """
        get used prefixes to be able to reinstall their routes
        """
        # query = "SELECT {0}.prefix FROM {0} INNER JOIN {1} ON {0}.prefix = {1}.prefix WHERE {0}.active = 1".format(self.table_prefixes, self.table_routes)
        query = f"SELECT {self.table_prefixes}.prefix FROM {self.table_prefixes} INNER JOIN {self.table_routes} ON " \
                f"{self.table_prefixes}.prefix = {self.table_routes}.prefix WHERE {self.table_prefixes}.active = 1"
        answer = self.query(query)
        active_prefixes = list()
        if answer is not None:
            for prefix in answer:
                active_prefixes.append(prefix[0])
        return active_prefixes

    def get_inactive_prefixes(self):
        """
        get unused prefixes to be able to delete their routes
        """
        # query = "SELECT {0}.prefix FROM {0} INNER JOIN {1} ON {0}.prefix = {1}.prefix WHERE {0}.active = 0".format(self.table_prefixes, self.table_routes)
        query = f"SELECT {self.table_prefixes}.prefix FROM {self.table_prefixes} INNER JOIN {self.table_routes} " \
                f"ON {self.table_prefixes}.prefix = {self.table_routes}.prefix WHERE {self.table_prefixes}.active = 0"
        answer = self.query(query)
        inactive_prefixes = list()
        if answer is not None:
            for prefix in answer:
                inactive_prefixes.append(prefix[0])
        return inactive_prefixes

    def get_route(self, prefix):
        """
        get all route parameters plus class for a certain prefix - mostly to delete the route
        """
        # query = "SELECT {0}.length, {0}.router, {1}.class FROM {0} INNER JOIN {1} WHERE {0}.prefix = {1}.prefix AND {1}.prefix = '{2}'".format(self.table_routes, self.table_prefixes, prefix)
        query = f"SELECT {self.table_routes}.length, {self.table_routes}.router, {self.table_prefixes}.class FROM " \
                f"{self.table_routes} INNER JOIN {self.table_prefixes} WHERE {self.table_routes}.prefix = " \
                f"{self.table_prefixes}.prefix AND {self.table_prefixes}.prefix = '{prefix}'"
        answer = self.query(query)
        if answer is not None:
            if len(answer) > 0:
                if len(answer[0]) > 0:
                    return answer[0]
                else:
                    # calling method expects triple of length, router and class - get None if nothing there
                    return None, None, None
            else:
                return None, None, None
        else:
            return None, None, None

    @clean_query_answer
    def release_lease(self, address, now):
        """
        release a lease via setting its active flag to False
        set last_message to 8 because of RELEASE messages having this message id
        """
        query = f"UPDATE {self.table_leases} SET active = 0, last_message = 8, last_update = '{now}' WHERE address = '{address}'"
        self.query(query)

    @clean_query_answer
    def release_prefix(self, prefix, now):
        """
        release a prefix via setting its active flag to False
        set last_message to 8 because of RELEASE messages having this message id
        """
        query = f"UPDATE {self.table_prefixes} SET active = 0, last_message = 8, last_update = '{now}' WHERE prefix = '{prefix}'"
        self.query(query)

    @clean_query_answer
    def check_number_of_leases(self, prefix='', range_from='', range_to=''):
        """
        check how many leases are stored - used to find out if address range has been exceeded
        """
        query = f"SELECT COUNT(address) FROM {self.table_leases} WHERE address LIKE '{prefix}%' AND " \
                f"'{prefix + range_from}' <= address AND address <= '{prefix + range_to}'"
        return self.query(query)

    @clean_query_answer
    def check_number_of_prefixes(self, prefix='', length='', range_from='', range_to=''):
        """
        check how many leases are stored - used to find out if address range has been exceeded
        """
        query = f"SELECT COUNT(prefix) FROM {self.table_prefixes} WHERE prefix LIKE '{prefix}%' AND " \
                f"'{prefix + range_from + ((128 - int(length)) // 4) * '0'}' <= prefix AND " \
                f"prefix <= '{prefix + range_to + ((128 - int(length)) // 4) * '0'}'"
        return self.query(query)

    def check_lease(self, address, transaction):
        """
        check state of a lease for REBIND and RENEW messages
        """
        # attributes to identify host and lease
        if cfg.IGNORE_IAID:
            query = f"SELECT DISTINCT hostname, address, type, category, ia_type, class, preferred_until " \
                    f"FROM {self.table_leases} WHERE active = 1 AND " \
                    f"address = '{address}' AND " \
                    f"mac = '{transaction.mac}' AND " \
                    f"duid = '{transaction.duid}'"
        else:
            query = f"SELECT DISTINCT hostname, address, type, category, ia_type, class, preferred_until " \
                    f"FROM {self.table_leases} WHERE active = 1 AND " \
                    f"address = '{address}' AND " \
                    f"mac = '{transaction.mac}' AND " \
                    f"duid = '{transaction.duid}' AND " \
                    f"iaid = '{transaction.iaid}'"
        return self.query(query)

    def check_prefix(self, prefix, length, transaction):
        """
        check state of a prefix for REBIND and RENEW messages
        """
        # attributes to identify host and lease
        if cfg.IGNORE_IAID:
            query = f"SELECT DISTINCT hostname, prefix, length, type, category, class, preferred_until " \
                    f"FROM {self.table_prefixes} WHERE active = 1 AND " \
                    f"prefix = '{prefix}' AND " \
                    f"length = '{length}' AND " \
                    f"mac = '{transaction.mac}' AND " \
                    f"duid = '{transaction.duid}'"
        else:
            query = f"SELECT DISTINCT hostname, prefix, length, type, category, class, preferred_until " \
                    f"FROM {self.table_prefixes} WHERE active = 1 AND " \
                    f"prefix = '{prefix}' AND " \
                    f"length = '{length}' AND " \
                    f"mac = '{transaction.mac}' AND " \
                    f"duid = '{transaction.duid}' AND " \
                    f"iaid = '{transaction.iaid}'"
        return self.query(query)

    def check_advertised_lease(self, transaction=None, category='', atype=''):
        """
        check if there are already advertised addresses for client
        """
        # attributes to identify host and lease
        if cfg.IGNORE_IAID:
            query = f"SELECT address FROM {self.table_leases} WHERE last_message = 1 AND " \
                    f"active = 1 AND " \
                    f"mac = '{transaction.mac}' AND " \
                    f"duid = '{transaction.duid}' AND " \
                    f"category = '{category}' AND " \
                    f"type = '{atype}'"
        else:
            query = f"SELECT address FROM {self.table_leases} WHERE last_message = 1 AND " \
                    f"active = 1 AND " \
                    f"mac = '{transaction.mac}' AND " \
                    f"duid = '{transaction.duid}' AND " \
                    f"iaid = '{transaction.iaid}' AND " \
                    f"category = '{category}' AND " \
                    f"type = '{atype}'"
        answer = self.query(query)
        if answer is not None:
            if len(answer) == 0:
                return False
            else:
                return answer[0][0]
        else:
            return False

    def check_advertised_prefix(self, transaction, category='', ptype=''):
        """
        check if there is already an advertised prefix for client
        """
        # attributes to identify host and lease
        if cfg.IGNORE_IAID:
            query = f"SELECT prefix, length FROM {self.table_prefixes} WHERE last_message = 1 AND " \
                    f"active = 1 AND " \
                    f"mac = '{transaction.mac}' AND " \
                    f"duid = '{transaction.duid}' AND " \
                    f"category = '{category}' AND " \
                    f"type = '{ptype}'"
        else:
            query = f"SELECT prefix, length FROM {self.table_prefixes} WHERE last_message = 1 AND " \
                    f"active = 1 AND " \
                    f"mac = '{transaction.mac}' AND " \
                    f"duid = '{transaction.duid}' AND " \
                    f"iaid = '{transaction.iaid}' AND " \
                    f"category = '{category}' AND " \
                    f"type = '{ptype}'"
        answer = self.query(query)
        if answer is not None:
            if len(answer) == 0:
                return False
            else:
                return answer[0][0]
        else:
            return False

    def release_free_leases(self, now):
        """
        release all invalid leases via setting their active flag to False
        """
        query = f"UPDATE {self.table_leases} SET active = 0, last_message = 0 WHERE valid_until < '{now}'"
        return self.query(query)

    def release_free_prefixes(self, now):
        """
            release all invalid prefixes via setting their active flag to False
        """
        query = f"UPDATE {self.table_prefixes} SET active = 0, last_message = 0 WHERE valid_until < '{now}'"
        return self.query(query)

    def remove_leases(self, now, category="random"):
        """
        remove all leases of a certain category like random - they will grow the database
        but be of no further use
        """
        query = f"DELETE FROM {self.table_leases} WHERE active = 0 AND " \
                f"category = '{category}' AND valid_until < '{now}'"
        return self.query(query)

    def remove_route(self, prefix):
        """
        remove a route which is not used anymore
        """
        query = f"DELETE FROM {self.table_routes} WHERE prefix = '{prefix}'"
        return self.query(query)

    def unlock_unused_advertised_leases(self, now):
        """
        unlock leases marked as advertised but apparently never been delivered
        let's say a client should have requested its formerly advertised address after 1 minute
        """
        query = f"UPDATE {self.table_leases} SET last_message = 0 WHERE last_message = 1 AND last_update < '{now + 60}'"
        return self.query(query)

    def unlock_unused_advertised_prefixes(self, now):
        """
            unlock prefixes marked as advertised but apparently never been delivered
            let's say a client should have requested its formerly advertised address after 1 minute
        """
        query = f"UPDATE {self.table_prefixes} SET last_message = 0 WHERE last_message = 1 AND " \
                f"last_update < '{now + 60}'"
        return self.query(query)

    def build_config_from_db(self, transaction):
        """
        get client config from db and build the appropriate config objects and indices
        """
        if transaction.client_config_dicts is None:
            # add client config which seems to fit to transaction
            transaction.client_config_dicts = ClientConfigDicts()

            if self.config_prefix_support:
                # 'mac LIKE' is necessary if multiple MACs are stored in config DB
                query = f"SELECT hostname, mac, duid, class, address, prefix, id FROM {self.table_hosts} WHERE " \
                        f"hostname = '{transaction.hostname}' OR " \
                        f"mac LIKE '%{transaction.mac}%' OR " \
                        f"duid = '{transaction.duid}'"
                answer = self.query(query)

                # read all sections of config file
                # a section here is a host
                # lowering MAC and DUID information in case they where upper in database
                for host in answer:
                    hostname, mac, duid, client_class, address, prefix, host_id = host
                    # lower some attributes to comply with values from request
                    if mac:
                        mac = listify_option(mac.lower())
                    if duid:
                        duid = duid.lower()
                    if address:
                        address = listify_option(address.lower())
                    if prefix:
                        prefix = listify_option(prefix.lower())

                    transaction.client_config_dicts.hosts[hostname] = ClientConfig(hostname=hostname,
                                                                                   mac=mac,
                                                                                   duid=duid,
                                                                                   client_class=client_class,
                                                                                   address=address,
                                                                                   prefix=prefix,
                                                                                   host_id=host_id)
                    # and put the host objects into index
                    if transaction.client_config_dicts.hosts[hostname].MAC:
                        for m in transaction.client_config_dicts.hosts[hostname].MAC:
                            if m not in transaction.client_config_dicts.index_mac:
                                transaction.client_config_dicts.index_mac[m] = [transaction.client_config_dicts.hosts[hostname]]
                            else:
                                transaction.client_config_dicts.index_mac[m]. \
                                    append(transaction.client_config_dicts.hosts[hostname])

                    # add DUIDs to IndexDUID
                    if transaction.client_config_dicts.hosts[hostname].DUID != '':
                        if transaction.client_config_dicts.hosts[hostname].DUID not in transaction.client_config_dicts.index_duid:
                            transaction.client_config_dicts.index_duid[transaction.client_config_dicts.hosts[hostname].DUID] = \
                                [transaction.client_config_dicts.hosts[hostname]]
                        else:
                            transaction.client_config_dicts.index_duid[transaction.client_config_dicts.hosts[hostname].DUID]. \
                                append(transaction.client_config_dicts.hosts[hostname])

                    # some cleaning
                    del host, mac, duid, address, prefix, client_class, host_id
            else:
                # 'mac LIKE' is necessary if multiple MACs are stored in config DB
                query = f"SELECT hostname, mac, duid, class, address, id FROM {self.table_hosts} WHERE " \
                        f"hostname = '{transaction.hostname}' OR " \
                        f"mac LIKE '%{transaction.mac}%' OR " \
                        f"duid = '{transaction.duid}'"
                answer = self.query(query)

                # read all sections of config file
                # a section here is a host
                # lowering MAC and DUID information in case they where upper in database
                for host in answer:
                    hostname, mac, duid, client_class, address, host_id = host
                    # lower some attributes to comply with values from request
                    if mac:
                        mac = listify_option(mac.lower())
                    if duid:
                        duid = duid.lower()
                    if address:
                        address = listify_option(address.lower())

                    transaction.client_config_dicts.hosts[hostname] = ClientConfig(hostname=hostname,
                                                                                   mac=mac,
                                                                                   duid=duid,
                                                                                   client_class=client_class,
                                                                                   address=address,
                                                                                   host_id=host_id)
                    # and put the host objects into index
                    if transaction.client_config_dicts.hosts[hostname].MAC:
                        for m in transaction.client_config_dicts.hosts[hostname].MAC:
                            if m not in transaction.client_config_dicts.index_mac:
                                transaction.client_config_dicts.index_mac[m] = [transaction.client_config_dicts.hosts[hostname]]
                            else:
                                transaction.client_config_dicts.index_mac[m]. \
                                    append(transaction.client_config_dicts.hosts[hostname])

                    # add DUIDs to IndexDUID
                    if transaction.client_config_dicts.hosts[hostname].DUID != '':
                        if transaction.client_config_dicts.hosts[hostname].DUID not in transaction.client_config_dicts.index_duid:
                            transaction.client_config_dicts.index_duid[transaction.client_config_dicts.hosts[hostname].DUID] = \
                                [transaction.client_config_dicts.hosts[hostname]]
                        else:
                            transaction.client_config_dicts.index_duid[transaction.client_config_dicts.hosts[hostname].DUID]. \
                                append(transaction.client_config_dicts.hosts[hostname])

                    # some cleaning
                    del host, mac, duid, address, client_class, host_id

    def get_client_config_by_mac(self, transaction):
        """
        get host and its information belonging to that mac
        """
        hosts = list()
        mac = transaction.mac

        if mac in transaction.client_config_dicts.index_mac:
            hosts.extend(transaction.client_config_dicts.index_mac[mac])
            return hosts
        else:
            return None

    def get_client_config_by_duid(self, transaction):
        """
            get host and its information belonging to that DUID
        """
        # get client config that most probably seems to fit
        hosts = list()
        duid = transaction.duid

        if duid in transaction.client_config_dicts.index_duid:
            hosts.extend(transaction.client_config_dicts.index_duid[duid])
            return hosts
        else:
            return None

    def get_client_config_by_hostname(self, transaction):
        """
            get host and its information by hostname
        """
        hostname = transaction.hostname
        if hostname in transaction.client_config_dicts.hosts:
            return [transaction.client_config_dicts.hosts[hostname]]
        else:
            return None

    def get_client_config(self, hostname='', client_class='', duid='', address=[], mac=[], host_id=''):
        """
            give back ClientConfig object
        """
        return ClientConfig(hostname=hostname,
                            client_class=client_class,
                            duid=duid,
                            address=address,
                            mac=mac,
                            host_id=host_id)

    def store_mac_llip(self, mac, link_local_ip, now):
        """
            store MAC-link-local-ip-mapping
        """
        query = f"SELECT mac FROM macs_llips WHERE mac='{mac}'"
        db_entry = self.query(query)
        # if known already update timestamp of MAC-link-local-ip-mapping
        if not db_entry or db_entry == []:
            query = f"INSERT INTO macs_llips (mac, link_local_ip, last_update) " \
                    f"VALUES ('{mac}', '{link_local_ip}', '{now}')"
            self.query(query)
        else:
            query = f"UPDATE macs_llips SET link_local_ip = '{link_local_ip}', last_update = '{now}' WHERE mac = '{mac}'"
            self.query(query)

    @clean_query_answer
    def get_dynamic_prefix(self):
        query = "SELECT item_value FROM meta WHERE item_key = 'dynamic_prefix'"
        return self.query(query)

    def store_dynamic_prefix(self, prefix):
        """
            store dynamic prefix to be persistent after restart of dhcpy6d
        """
        query = "SELECT item_value FROM meta WHERE item_key = 'dynamic_prefix'"
        db_entry = self.query(query)

        # if already existing just update dynamic prefix
        if not db_entry or db_entry == []:
            query = f"INSERT INTO meta (item_key, item_value) VALUES ('dynamic_prefix', '{prefix}')"
            self.query(query)
        else:
            query = f"UPDATE meta SET item_value = '{prefix}' WHERE item_key = 'dynamic_prefix'"
            self.query(query)

    def collect_macs_from_db(self):
        """
            collect all known MACs and link local addresses from database at startup
            to reduce attempts to read neighbor cache
        """
        query = f'SELECT link_local_ip, mac, last_update FROM {self.table_macs_llips}'
        answer = self.query(query)
        if answer:
            for m in answer:
                try:
                    # m[0] is LLIP, m[1] is the matching MAC
                    # interface is ignored
                    collected_macs[m[0]] = NeighborCacheRecord(llip=m[0], mac=m[1], now=m[2])
                except Exception as err:
                    print(err)
                    traceback.print_exc(file=sys.stdout)
                    sys.stdout.flush()
                    return None

    def db_query(self, query):
        """
            no not execute query on DB - dummy
        """
        # return empty tuple as dummy
        return ()


class DB(Store):
    """
        MySQL and PostgreSQL database interface
        for robustness http://stackoverflow.com/questions/207981/how-to-enable-mysql-client-auto-re-connect-with-mysqldb
    """
    connection = False
    cursor = False

    def __init__(self, query_queue, answer_queue):
        Store.__init__(self, query_queue, answer_queue)
        self.connection = None
        try:
            self.db_connect()
        except Exception as err:
            print(err)

    def db_connect(self):
        """
        Connect to database server according to database type
        """
        pass



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from binascii import (hexlify,
                      unhexlify)
import shlex
import socket
import sys

# whitespace for options with more than one value
WHITESPACE = ' ,'

# define address characters once - for decompress_ipv6
ADDRESS_CHARS_STRICT = ':0123456789abcdef'
ADDRESS_CHARS_NON_STRICT = ':0123456789abcdefx'

# localhost
LOCALHOST = '::1'
LOCALHOST_LLIP = '00000000000000000000000000000001'
LOCALHOST_INTERFACES = ['', 'lo', 'lo0']


class Interface:
    """
    hold interface information
    interface information comes in tuple from socket.if_nameindex()
    """

    def __init__(self, interface_tuple):
        self.index, self.name = interface_tuple


class NeighborCacheRecord:
    """
    object for neighbor cache entries to be returned by get_neighbor_cache_linux() and in CollectedMACs
    .interface is only interesting for real neighbor cache records, to be ignored for collected MACs stored in DB
    """

    def __init__(self, llip='', mac='', interface='', now=0):
        self.llip = llip
        self.mac = mac
        self.interface = interface
        self.timestamp = now


def convert_dns_to_binary(name):
    """
    convert domain name as described in RFC 1035, 3.1
    """
    binary = ''
    domain_parts = name.split('.')
    for domain_part in domain_parts:
        binary += f'{len(domain_part):02x}'  # length of Domain Name Segments
        binary += hexlify(domain_part.encode()).decode()
    # final zero size octet following RFC 1035
    binary += '00'
    return binary


def convert_binary_to_dns(binary):
    """
    convert domain name from hex like in RFC 1035, 3.1
    """
    name = ''
    binary_parts = binary
    while len(binary_parts) > 0:
        # RFC 1035 - domain names are sequences of labels separated by length octets
        length = int(binary_parts[0:2], 16)
        # lenght*2 because 2 charse represent a byte
        label = unhexlify(binary_parts[2:2 + length * 2]).decode()
        binary_parts = binary_parts[2 + length * 2:]
        name += label
        # insert '.' if this is not the last label of FQDN
        # >2 because last byte is the zero byte terminator
        if len(binary_parts) > 2:
            name += '.'
    return str(name)


def build_option(number, payload):
    """
    glue option with payload
    """
    # option number and length take 2 byte each so the string has to be 4 chars long
    option = f'{number:04x}'  # option number
    option += f'{len(payload) // 2:04x}'  # payload length, /2 because 2 chars are 1 byte
    option += payload
    return option


def correct_mac(mac):
    """
    OpenBSD shortens MAC addresses in ndp output - here they grow again
    """
    decompressed = [f'{(int(m, 16)):02x}' for m in mac.split(':')]
    return ':'.join(decompressed)


def colonify_mac(mac):
    """
    return complete MAC address with colons
    """
    if type(mac) == bytes:
        mac = mac.decode()
    return ':'.join((mac[0:2], mac[2:4], mac[4:6], mac[6:8], mac[8:10], mac[10:12]))


def decompress_ip6(ip6, strict=True):
    """
    decompresses shortened IPv6 address and returns it as ':'-less 32 character string
    additionally allows testing for prototype address with less strict set of allowed characters
    """

    ip6 = ip6.lower()
    # cache some repeated calls
    colon_count1 = ip6.count(':')
    colon_count2 = ip6.count('::')
    colon_count3 = ip6.count(':::')

    # if in strict mode there are no hex numbers and ':' something is wrong
    if strict:
        for c in ip6:
            if c not in ADDRESS_CHARS_STRICT:
                raise Exception(f'{ip6} should consist only of : 0 1 2 3 4 5 6 7 8 9 a b c d e f')
    else:
        # used for comparison of leases with address pattern - X replace the dynamic part of the address
        for c in ip6:
            if c not in ADDRESS_CHARS_NON_STRICT:
                raise Exception(f'{ip6} should consist only of : 0 1 2 3 4 5 6 7 8 9 a b c d e f x')
    # nothing to do
    if len(ip6) == 32 and colon_count1 == 0:
        return ip6

    # larger heaps of :: smell like something wrong
    if colon_count2 > 1 or colon_count3 >= 1:
        raise Exception(f"{ip6} has too many accumulated ':'")

    # less than 7 ':' but no '::' also make a bad impression
    if colon_count1 < 7 and colon_count2 != 1:
        raise Exception(f"{ip6} is missing some ':'")

    # replace :: with :0000:: - the last ':' will be cut of finally
    while ip6.count(':') < 8 and ip6.count('::') == 1:
        ip6 = ip6.replace('::', ':0000::')

    # remaining ':' will be cut off
    ip6 = ip6.replace('::', ':')

    # ':' at the beginning have to be filled up with 0000 too
    if ip6.startswith(':'):
        ip6 = '0000' + ip6

    # if a segment is shorter than 4 chars the gaps get filled with zeros
    ip6_segments_source = ip6.split(':')
    ip6_segments_target = list()
    for s in ip6_segments_source:
        if len(s) > 4:
            raise Exception(f"{ip6} has segment with more than 4 digits")
        else:
            ip6_segments_target.append(s.zfill(4))

    # return with separator (mostly '')
    return ''.join(ip6_segments_target)


def colonify_ip6(address):
    """
    return complete IPv6 address with colons
    """
    if address:
        if type(address) == bytes:
            address = address.decode()
        return ':'.join((address[0:4], address[4:8], address[8:12], address[12:16],
                         address[16:20], address[20:24], address[24:28], address[28:32]))
    else:
        # return 'n/a'
        # provoke crash to see what happens with un-addresses
        print('Uncolonifyable address:', address)
        return False


def convert_prefix_inline(prefix):
    """
    check if supplied prefix is valid and convert it
    """
    address, length = split_prefix(prefix)
    address = decompress_ip6(address)
    return {"address": address,
            "length": length}


def combine_prefix_length(prefix, length):
    """
    add prefix and length to 'prefix/length' string
    """
    return f'{prefix}/{length}'


def split_prefix(prefix):
    """
    split prefix and length from 'prefix/length' notation
    """
    return prefix.split('/')


def decompress_prefix(prefix, length):
    """
    return prefix with decompressed address part
    """
    return combine_prefix_length(decompress_ip6(prefix), length)


def error_exit(message='An error occured.', status=1):
    """
    exit with given error message
    allow prefix, especially for spitting out section of configuration errors
    """
    sys.stderr.write(f'\n{message}\n\n')
    sys.exit(status)


def listify_option(option):
    """
    return any comma or space separated option as list
    if erroneously a list has been given return it unchanged
    """
    if option:
        if type(option) == str:
            lex = shlex.shlex(option)
            lex.whitespace = WHITESPACE
            lex.wordchars += ':.-/'
            return list(lex)
        elif type(option) == list:
            return(option)
        else:
            return False
    else:
        return None


def send_control_message(message):
    """
    Send a control message to the locally running dhcpy6d daemon
    """
    # clean message of quotations marks
    message = message.strip('"').encode('utf8')
    socket_control = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    socket_control.sendto(message, ('::1', 547))


def convert_mac_to_eui64(mac):
    """
    Convert a MAC address to a EUI64 address
    """
    # http://tools.ietf.org/html/rfc4291#section-2.5.1
    # only ':' come in MACs from get_neighbor_cache_linux()
    eui64 = mac.replace(':', '')
    eui64 = eui64[0:6] + 'fffe' + eui64[6:]
    eui64 = hex(int(eui64[0:2], 16) ^ 2)[2:].zfill(2) + eui64[2:]

    split_string = lambda x, n: [x[i:i + n] for i in range(0, len(x), n)]

    return ':'.join(split_string(eui64, 4))


def get_interfaces():
    """
    return dict full of Interface objects
    :return:
    """
    interfaces = {}
    for interface_tuple in socket.if_nameindex():
        interface = Interface(interface_tuple)
        interfaces[interface.name] = interface
    return interfaces



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from binascii import hexlify

from dhcpy6d.client import Client
from dhcpy6d.helpers import build_option
from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    Option 59 Network Boot
    https://tools.ietf.org/html/rfc5970
    """
    def build(self, transaction=None, **kwargs):
        # dummy empty defaults
        response_string_part = ''
        options_answer_part = None

        # build client if not done yet
        if transaction.client is None:
            transaction.client = Client(transaction)

        bootfiles = transaction.client.bootfiles
        if len(bootfiles) > 0:
            # TODO add preference logic
            bootfile_url = bootfiles[0].BOOTFILE_URL
            transaction.client.chosen_boot_file = bootfile_url
            bootfile_options = hexlify(bootfile_url).decode()
            response_string_part += build_option(self.number, bootfile_options)
            # options in answer to be logged
            options_answer_part = self.number

        return response_string_part, options_answer_part



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from grp import getgrnam
import logging
from logging import (Formatter,
                     getLogger,
                     StreamHandler)
from logging.handlers import (SysLogHandler,
                              WatchedFileHandler)
from os import chown
from pwd import getpwnam
from socket import gethostname

from .config import cfg

# globally available logging instace
log = getLogger('dhcpy6d')

if cfg.LOG:
    formatter = Formatter('{asctime} {name} {levelname} {message}', style='{')
    log.setLevel(logging.__dict__[cfg.LOG_LEVEL])
    if cfg.LOG_FILE != '':
        chown(cfg.LOG_FILE, getpwnam(cfg.USER).pw_uid, getgrnam(cfg.GROUP).gr_gid)
        log_handler = WatchedFileHandler(cfg.LOG_FILE)
        log_handler.setFormatter(formatter)
        log.addHandler(log_handler)
    # std err console output
    if cfg.LOG_CONSOLE:
        log_handler = StreamHandler()
        log_handler.setFormatter(formatter)
        log.addHandler(log_handler)
    if cfg.LOG_SYSLOG:
        # time should be added by syslog daemon
        hostname = gethostname().split('.')[0]
        formatter = Formatter(hostname + ' {name} {levelname} {message}', style='{')
        # if /socket/file is given use this as address
        if cfg.LOG_SYSLOG_DESTINATION.startswith('/'):
            destination = cfg.LOG_SYSLOG_DESTINATION
        # if host and port are defined use them...
        elif cfg.LOG_SYSLOG_DESTINATION.count(':') == 1:
            destination = tuple(cfg.LOG_SYSLOG_DESTINATION.split(':'))
        # ...otherwise add port 514 to given host address
        else:
            destination = (cfg.LOG_SYSLOG_DESTINATION, 514)
        log_handler = SysLogHandler(address=destination,
                                    facility=SysLogHandler.__dict__['LOG_' + cfg.LOG_SYSLOG_FACILITY])
        log_handler.setFormatter(formatter)
        log.addHandler(log_handler)



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import random

from ..config import cfg
from ..helpers import (convert_mac_to_eui64,
                       decompress_ip6)
from ..log import log
from ..storage import volatile_store


def parse_pattern_address(address, client_config, transaction):
    """
        parse address pattern and replace variables with current values
    """
    # parse all pattern parts
    a = address.PATTERN

    # if dhcpy6d got a new (mostly dynamic) prefix at start insert it here
    if cfg.PREFIX is not None:
        a = a.replace('$prefix$', cfg.PREFIX)

    # check different client address categories - to be extended!
    if address.CATEGORY == 'mac':
        macraw = ''.join(transaction.mac.split(':'))
        a = a.replace('$mac$', ':'.join((macraw[0:4], macraw[4:8], macraw[8:12])))
    elif address.CATEGORY == 'eui64':
        # https://tools.ietf.org/html/rfc4291#section-2.5.1
        mac = transaction.mac
        a = a.replace('$eui64$', convert_mac_to_eui64(mac))
    elif address.CATEGORY in ['fixed', 'dns']:
        # No patterns for fixed address, let's bail
        return None
    elif address.CATEGORY == 'id':
        # if there is an ID build address
        if str(client_config.ID) != '':
            a = a.replace('$id$', str(client_config.ID))
        else:
            return None
    elif address.CATEGORY == 'random':
        # first check if address already has been advertised
        advertised_address = volatile_store.check_advertised_lease(transaction,
                                                                   category='random',
                                                                   atype=address.TYPE)
        # when address already has been advertised for this client use it
        if advertised_address:
            a = advertised_address
        else:
            ra = str(hex(random.getrandbits(64)))[2:][:-1]
            ra = ':'.join((ra[0:4], ra[4:8], ra[8:12], ra[12:16]))
            # subject to change....
            a = a.replace('$random64$', ra)
    elif address.CATEGORY == 'range':
        range_from, range_to = address.RANGE.split('-')
        if len(range_from) < 4:
            range_from = '0' * (4-len(range_from)) + range_from
        if len(range_to) < 4:
            range_to = '0' * (4-len(range_to)) + range_to
        if range_from > range_to:
            range_from, range_to = range_to, range_from

        # expecting range-range at the last octet, 'prefix' means the first seven octets here
        # - is just shorter than the_first_seven_octets
        prefix = decompress_ip6(a.replace('$range$', '0000'))[:28]

        # the following steps are done to find a collision-free lease in given range
        # check if address already has been advertised - important for REPLY after SOLICIT-ADVERTISE-REQUEST
        advertised_address = volatile_store.check_advertised_lease(transaction, category='range', atype=address.TYPE)
        # when address already has been advertised for this client use it
        if advertised_address:
            a = advertised_address
        else:
            # check if requesting client still has an active lease that could be reused
            lease = volatile_store.get_range_lease_for_recycling(prefix=prefix,
                                                                 range_from=range_from,
                                                                 range_to=range_to,
                                                                 duid=transaction.duid,
                                                                 mac=transaction.mac)
            # the found lease has to be in range - important after changed range boundaries
            if lease is not None and range_from <= lease[28:].lower() <= range_to:
                a = ':'.join((lease[0:4], lease[4:8], lease[8:12], lease[12:16],
                              lease[16:20], lease[20:24], lease[24:28], lease[28:32]))
            else:
                # get highest active lease to increment address about 1
                lease = volatile_store.get_highest_range_lease(prefix=prefix, range_from=range_from, range_to=range_to)
                # check if highest active lease still fits into range
                if lease is not None and range_from <= lease[28:].lower() < range_to:
                    # if highest lease + 1 would not fit range limit is reached
                    if lease[28:].lower() >= range_to:
                        # try to get one of the inactive old leases
                        lease = volatile_store.get_oldest_inactive_range_lease(prefix=prefix,
                                                                               range_from=range_from,
                                                                               range_to=range_to)
                        if lease is None:
                            # if none is available limit is reached and nothing returned
                            log.critical(f'Address space {prefix}[{range_from}-{range_to}] exceeded')
                            return None
                        else:
                            # if lease is OK use it
                            a = lease
                    else:
                        # otherwise increase current maximum range limit by 1
                        a = a.replace('$range$', str(hex(int(lease[28:], 16) + 1)).split('x')[1])
                else:
                    # if there is no lease yet or range limit is reached try to reactivate an old inactive lease
                    lease = volatile_store.get_oldest_inactive_range_lease(prefix=prefix,
                                                                           range_from=range_from,
                                                                           range_to=range_to)
                    if lease is None:
                        # if there are no leases stored yet initiate lease storage
                        # this will be done only once - the first time if there is no other lease yet
                        # so it is safe to start from range_from
                        if volatile_store.check_number_of_leases(prefix, range_from, range_to) <= 1:
                            a = a.replace('$range$', range_from)
                        else:
                            # if none is available limit is reached and nothing returned
                            log.critical(f'Address space {prefix}[{range_from}-{range_to}] exceeded')
                            return None
                    else:
                        # if there is a lease it might be used
                        a = lease

    return decompress_ip6(a)


def parse_pattern_prefix(pattern, client_config, transaction):
    """
        parse address pattern and replace variables with current values
    """
    # parse all pattern parts
    p = pattern.PATTERN

    # if dhcpy6d got a new (mostly dynamic) prefix at start insert it here
    p = p.replace('$prefix$', cfg.PREFIX)

    if pattern.CATEGORY == 'id':
        # if there is an ID build address
        if str(client_config.ID) != '':
            p = p.replace('$id$', str(client_config.ID))
        else:
            return None

    elif pattern.CATEGORY == 'range':
        range_from, range_to = pattern.RANGE.split('-')
        if len(range_from) < 4:
            range_from = '0' * (4-len(range_from)) + range_from
        if len(range_to) < 4:
            range_to = '0' * (4-len(range_to)) + range_to
        if range_from > range_to:
            range_from, range_to = range_to, range_from

        # contrary to addresses the prefix $range$ part of the pattern is expected
        # somewhere at the left part of the pattern
        # here the 128 Bit sum up to 32 characters in address/prefix string so prefix_range_index has to be calculated
        # as first character of range part of prefix - assuming steps of width 4
        prefix_range_index = int(pattern.LENGTH)//4-4
        # prefix itself has a prefix - the first part of the prefix pattern
        prefix_prefix = decompress_ip6(p.replace('$range$', '0000'))[:prefix_range_index]

        # the following steps are done to find a collision-free lease in given range
        # check if address already has been advertised - important for REPLY after SOLICIT-ADVERTISE-REQUEST
        advertised_prefix = volatile_store.check_advertised_prefix(transaction, category='range', ptype=pattern.TYPE)

        # when address already has been advertised for this client use it
        if advertised_prefix:
            p = advertised_prefix
        else:
            # check if requesting client still has an active prefix that could be reused
            prefix = volatile_store.get_range_prefix_for_recycling(prefix=prefix_prefix,
                                                                   length=pattern.LENGTH,
                                                                   range_from=range_from,
                                                                   range_to=range_to,
                                                                   duid=transaction.duid,
                                                                   mac=transaction.mac)
            # the found prefix has to be in range - important after changed range boundaries
            if prefix is not None:
                if range_from <= prefix[prefix_range_index:prefix_range_index+4].lower() <= range_to:
                    p = ':'.join((prefix[0:4], prefix[4:8], prefix[8:12], prefix[12:16],
                                  prefix[16:20], prefix[20:24], prefix[24:28], prefix[28:32]))
                else:
                    # if prefixes are exceeded or something went wrong with from/to ranges return none
                    log.critical('Prefix address space %s[%s-%s] exceeded or something is wrong with from/to ranges' %
                                 (prefix_prefix, range_from, range_to))
                    return None
            else:
                # get highest active lease to increment address about 1
                prefix = volatile_store.get_highest_range_prefix(prefix=prefix_prefix,
                                                                 length=pattern.LENGTH,
                                                                 range_from=range_from,
                                                                 range_to=range_to)
                # check if highest active lease still fits into range
                if prefix is not None:
                    if range_from <= prefix[prefix_range_index:prefix_range_index+4].lower() < range_to:
                        # if highest lease + 1 would not fit range limit is reached
                        if prefix[prefix_range_index:prefix_range_index+4].lower() >= range_to:
                            # try to get one of the inactive old leases
                            prefix = volatile_store.get_oldest_inactive_range_prefix(prefix=prefix_prefix,
                                                                                     length=pattern.LENGTH,
                                                                                     range_from=range_from,
                                                                                     range_to=range_to)
                            if prefix is None:
                                # if none is available limit is reached and nothing returned
                                log.critical(f'Prefix address space {prefix_prefix}[{range_from}-{range_to}] exceeded')
                                return None
                            else:
                                # if lease is OK use it
                                p = prefix
                        else:
                            # otherwise increase current maximum range limit by 1
                            p = p.replace('$range$',
                                          str(hex(int(prefix[prefix_range_index:prefix_range_index+4].lower(), 16)
                                                  + 1)).split('x')[1])

                    else:
                        # if there is no lease yet or range limit is reached try to reactivate an old inactive lease
                        prefix = volatile_store.get_oldest_inactive_range_prefix(prefix=prefix_prefix,
                                                                                 length=pattern.LENGTH,
                                                                                 range_from=range_from,
                                                                                 range_to=range_to)
                        if prefix is None:
                            # if there are no leases stored yet initiate lease storage
                            # this will be done only once - the first time if there is no other lease yet
                            # so it is safe to start from range_from
                            if volatile_store.check_number_of_prefixes(prefix=prefix_prefix,
                                                                       length=pattern.LENGTH,
                                                                       range_from=range_from,
                                                                       range_to=range_to) <= 1:
                                p = p.replace('$range$', range_from)
                            else:
                                # if none is available limit is reached and nothing returned
                                log.critical(
                                    f'Prefix address space {prefix_prefix}[{range_from}-{range_to}] exceeded')
                                return None
                        else:
                            # if there is a lease it might be used
                            p = prefix

                else:
                    # if there is no lease yet or range limit is reached try to reactivate an old inactive lease
                    prefix = volatile_store.get_oldest_inactive_range_prefix(prefix=prefix_prefix,
                                                                             length=pattern.LENGTH,
                                                                             range_from=range_from,
                                                                             range_to=range_to)
                    if prefix is None:
                        # if there are no leases stored yet initiate lease storage
                        # this will be done only once - the first time if there is no other lease yet
                        # so it is safe to start from range_from
                        if volatile_store.check_number_of_prefixes(prefix=prefix_prefix,
                                                                   length=pattern.LENGTH,
                                                                   range_from=range_from,
                                                                   range_to=range_to) <= 1:
                            p = p.replace('$range$', range_from)
                        else:
                            # if none is available limit is reached and nothing returned
                            log.critical(f'Prefix address space {prefix_prefix}[{range_from}-{range_to}] exceeded')
                            return None
                    else:
                        # if there is a lease it might be used
                        p = prefix

    return decompress_ip6(p)



# DHCPy6d DHCPv6 Daemon
#
# Copyright (C) 2009-2022 Henri Wahl <henri@dhcpy6d.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from dhcpy6d.options import OptionTemplate


class Option(OptionTemplate):
    """
    Option 6 Option Request Option
    """
    def initialize(self, transaction=None, option=None, **kwargs):
        options_request = []
        options = option[:]
        # cut given option (which contains all requested options) into pieces
        while len(options) > 0:
            options_request.append(int(options[0:4], 16))
            options = options[4:]
        transaction.options_request = options_request
